from dataclasses import dataclass, field
from typing import List

from ..utils.Base import Base
from ..utils.DictDot import DictDot
from . import DomoAuth as dmda
from .routes import user_routes


class DomoUsers:
    def __init__(self, user):
        self.user = user

    @staticmethod
    def util_match_domo_users_to_emails(domo_users: list, user_email_list: List[str]) -> list:
        """pass in an array of user emails to match against an array of Domo User"""

        matches = []
        for idx, email in enumerate(user_email_list):
            matchUsr = next((domo_usr for domo_usr in domo_users if email.lower() == domo_usr.email_address.lower()),
                            None)
            if matchUsr is not None:
                matches.append(matchUsr)
        return matches

    def _users_to_domo_user(json_list, full_auth: dmda.DomoFullAuth):
        domo_users = []
        for json_obj in json_list:
            user = DomoUser._from_search_json(
                full_auth=full_auth,
                json_obj=json_obj)
            domo_users.append(user)

        return domo_users

    def _users_to_virtual_user(json_list, full_auth: dmda.DomoFullAuth):
        domo_users = []
        for json_obj in json_list:
            user = DomoUser._from_virtual_json(
                full_auth=full_auth,
                json_obj=json_obj)
            domo_users.append(user)

        return domo_users

    @classmethod
    async def all_users(cls, full_auth: dmda.DomoFullAuth, log_results: bool = False):

        res = await user_routes.get_all_users(full_auth=full_auth, log_results=log_results)

        if res.status == 200:
            json_list = res.response
            return cls._users_to_domo_user(json_list=json_list, full_auth=full_auth)

    @classmethod
    async def by_id(cls, user_ids: list[str],
                    full_auth: dmda.DomoFullAuth,
                    only_allow_one: bool = True,
                    log_results: bool = False, debug: bool = False) -> list:

        body = user_routes.generate_search_users_body_by_id(user_ids)
        res = await user_routes.search_users(full_auth=full_auth, body=body, log_results=log_results, debug=debug)

        if res.status == 200:
            json_list = res.response.get('users')

            if not json_list:
                return None

            domo_users = cls._users_to_domo_user(
                json_list, full_auth=full_auth)

            if only_allow_one:
                return domo_users[0]

            return domo_users

    @classmethod
    async def by_email(cls, email_address: str, full_auth: dmda.DomoFullAuth,
                       only_allow_one: bool = True,
                       log_results: bool = False, debug: bool = False) -> list:

        body = user_routes.generate_search_users_body_by_email(email_address)
        res = await user_routes.search_users(full_auth=full_auth, body=body, log_results=log_results, debug=debug)

        if res.status == 200:
            json_list = res.response.get('users')

            if not json_list:
                return None

            domo_users = cls._users_to_domo_user(
                json_list, full_auth=full_auth)

            if only_allow_one:
                domo_users = cls.util_match_domo_users_to_emails(
                    domo_users, [email_address])
                return domo_users[0]

            return domo_users

        return None

    @classmethod
    async def virtual_user_by_subscriber_instance(cls,
                                                  subscriber_instance: str,
                                                  full_auth: dmda.DomoFullAuth,
                                                  debug: bool = False, log_results: bool = False
                                                  ):
        res = await user_routes.search_virtual_user_by_subscriber_instance(full_auth=full_auth,
                                                                           subscriber_instance=subscriber_instance,
                                                                           log_results=log_results, debug=debug)

        if res.status != 200:
            return None

        json_list = res.response
        
        if debug:
            print (json_list)

        if not json_list:
            return None

        domo_users = cls._users_to_virtual_user(json_list, full_auth=full_auth)
        return domo_users[0]


@dataclass
class DomoUser(Base):
    id: str
    domo_instance: str
    full_auth: dmda.DomoFullAuth = field(repr=False)
    display_name: str = None
    email_address: str = None

    publisher_domain: str = None
    subscriber_domain: str = None
    virtual_user_id: str = None

    def __post_init__(self):
        Base().__init__()

    @classmethod
    def _from_search_json(cls, full_auth, json_obj):
        dd = DictDot(json_obj)

        u = cls(
            domo_instance=full_auth.domo_instance,
            full_auth=full_auth,
            id=dd.id or dd.userId,
            display_name=dd.displayName,
            email_address=dd.emailAddress
        )
        return u

    @classmethod
    def _from_virtual_json(cls, full_auth, json_obj):
        dd = DictDot(json_obj)

        u = cls(
            domo_instance=full_auth.domo_instance,
            full_auth=full_auth,
            id=dd.id,
            publisher_domain=dd.publisherDomain,
            subscriber_domain=dd.subscriberDomain,
            virtual_user_id=dd.virtualUserId
        )
        return u

    async def reset_password(self, new_password: str, debug: bool = False, log_results: bool = False):

        res = await user_routes.reset_password(full_auth=self.full_auth, user_id=self.id, new_password=new_password,
                                               debug=debug, log_results=log_results)

        return res

    @classmethod
    async def request_password_reset(cls, domo_instance: str, email: str, locale: str = 'en-us', debug: bool = False):

        return await user_routes.request_password_reset(domo_instance=domo_instance, email=email, locale=locale,
                                                        debug=debug)

    @classmethod
    async def create_user(cls, full_auth: dmda.DomoFullAuth, display_name, email, role_id, password: str = None,
                          send_password_reset_email: bool = False,
                          debug: bool = False, log_results: bool = False):

        res = await user_routes.create_user(full_auth=full_auth, display_name=display_name, email=email,
                                            role_id=role_id, debug=debug, log_results=log_results)

        if debug:
            print(res)

        if res.status != 200:
            return None

        dd = DictDot(res.response)
        u = cls(domo_instance=full_auth.domo_instance,
                full_auth=full_auth,
                id=dd.id or dd.userId,
                display_name=dd.displayName,
                email_address=dd.emailAddress)

        if password:
            await u.reset_password(new_password=password)

        if send_password_reset_email:
            await u.request_password_reset(domo_instance=full_auth.domo_instance,
                                           email=u.email_address)

        return u
    
    async def set_user_landing_page(self, 
                                    page_id:str,
                                    user_id: str = None,
                                    full_auth : dmda.DomoFullAuth = None,
                                   debug:bool = False):

        res = await user_routes.set_user_landing_page(full_auth = full_auth or self.full_auth, page_id = page_id, 
                                                      user_id = self.id or user_id , debug = debug)

        if res.status != 200:
            return False

        return True
            



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/account.ipynb.

# %% auto 0
__all__ = ['get_accounts', 'get_account_from_id', 'AccountConfig_InvalidDataProvider', 'get_account_config',
           'update_account_config', 'update_account_name', 'create_account', 'delete_account']

# %% ../../nbs/routes/account.ipynb 3
from typing import Union
import aiohttp
import httpx

import domolibrary.client.get_data as gd
import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.DomoAuth as dmda

# %% ../../nbs/routes/account.ipynb 4
async def get_accounts(auth: dmda.DomoAuth,
                       debug_api: bool = False, 
                       session: Union[aiohttp.ClientSession, httpx.AsyncClient, None] = None):
    """retrieve a list of all the accounts the user has read access to.  Note users with "Manage all accounts" will retrieve all account objects"""
    
    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts"

    return await gd.get_data(
        auth=auth,
        url=url,
        method='GET',
        debug_api=debug_api,
        session=session
    )

# %% ../../nbs/routes/account.ipynb 7
async def get_account_from_id(auth: dmda.DomoAuth, account_id: int,
                              debug_api: bool = False, session: aiohttp.ClientSession = None):
    """retrieves metadata about an account"""

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts/{account_id}?unmask=true"

    if debug_api:
        print(url)

    return await gd.get_data(
        auth=auth,
        url=url,
        method='GET',
        debug_api=debug_api,
        session=session
    )

# %% ../../nbs/routes/account.ipynb 11
class AccountConfig_InvalidDataProvider(Exception):
    def __init__(self, account_id:str, data_provider_type:str, domo_instance: str):
        message = f"Account - {account_id}, could not be retrieved with data_provider_type, '{data_provider_type}' from {domo_instance}"
        super().__init__( message)

async def get_account_config(auth: dmda.DomoAuth,
                             account_id: int,
                             data_provider_type: str ,
                             debug_api: bool = False, 
                             session: Union[aiohttp.ClientSession, httpx.AsyncClient, None] = None):

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/providers/{data_provider_type}/account/{account_id}?unmask=true"

    if debug_api:
        print(url)

    res = await gd.get_data(
        auth=auth,
        url=url,
        method='GET',
        debug_api=debug_api,
        session=session
    )

    if res.response == {}:
        raise AccountConfig_InvalidDataProvider(account_id= account_id, data_provider_type= data_provider_type, domo_instance=auth.domo_instance)
    
    return res

# %% ../../nbs/routes/account.ipynb 14
async def update_account_config(auth: dmda.DomoAuth,
                                account_id: int,
                                config_body: dict,
                                data_provider_type: str,
                                debug_api: bool = False, 
                                session: aiohttp.ClientSession = None):

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/providers/{data_provider_type}/account/{account_id}"

    if debug_api:
        print(url)

    return await gd.get_data(
        auth=auth,
        url=url,
        method='PUT',
        body=config_body,
        debug_api=debug_api,
        session=session
    )

# %% ../../nbs/routes/account.ipynb 15
async def update_account_name(auth: dmda.DomoAuth,
                              account_id: int,
                              account_name: str,
                              debug_api: bool = False, 
                              session: aiohttp.ClientSession = None):

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts/{account_id}/name"
    
    if debug_api:
        print(url)

    return await gd.get_data(
        auth=auth,
        url=url,
        method='PUT',
        body=account_name,
        content_type = "text/plain",
        debug_api=debug_api,
        session=session
    )

# %% ../../nbs/routes/account.ipynb 16
async def create_account(auth:dmda.DomoAuth, config_body:dict,
                         debug_api: bool = False, session: aiohttp.ClientSession = None):

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts"

    if debug_api:
        print(url)

    return await gd.get_data(
        auth=auth,
        url=url,
        method='POST',
        body = config_body,
        debug_api=debug_api,
        session=session
    )

# %% ../../nbs/routes/account.ipynb 17
async def delete_account(auth:dmda.DomoAuth,
                         account_id: str,
                         debug_api: bool = False, 
                         session: aiohttp.ClientSession = None):
    
    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts/{account_id}"

    if debug_api:
        print(url)

    return await gd.get_data(
        auth=auth,
        url=url,
        method='DELETE',
        debug_api=debug_api,
        session=session
    )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoPage.ipynb.

# %% auto 0
__all__ = ['DomoPage']

# %% ../../nbs/classes/50_DomoPage.ipynb 2
from fastcore.basics import patch_to

# %% ../../nbs/classes/50_DomoPage.ipynb 3
from dataclasses import dataclass, field

# import asyncio
# import aiohttp

import domolibrary.client.DomoAuth as dmda
import domolibrary.classes.DomoUser as dmdu
import domolibrary.utils.DictDot as util_dd

# from .routes import page_routes
# from ..utils.Base import Base

# %% ../../nbs/classes/50_DomoPage.ipynb 4
@dataclass
class DomoPage:
    id: str
    title: str = None
    parent_page_id: str = None
    auth: dmda.DomoAuth = field(default = None , repr = False)
    owners: list = field(default_factory=list)
    cards: list = field(default_factory=list)
    collections: list = field(default_factory=list)
    children: list = field(default_factory=list)

    def display_url(self):
        return f"https://{self.auth.domo_instance}.domo.com/page/{self.id}"

# %% ../../nbs/classes/50_DomoPage.ipynb 5
@patch_to(DomoPage, cls_method=True)
def _from_bootstrap(cls: DomoPage, page_obj, auth: dmda.DomoAuth = None):

    dd = page_obj
    if isinstance(page_obj, dict):
        dd = util_dd.DictDot(page_obj)

    domo_page = cls(id=dd.id, title=dd.title, auth=auth)

    if isinstance(dd.owners, list) and len(dd.owners) > 0:
        domo_page.owners = [
            dmdu.DomoUser._from_bootstrap_json(auth=auth, user_obj=user_dd)
            for user_dd in dd.owners
            if user_dd.type == 'USER'
        ]

        [print(other_dd) for other_dd in dd.owners
         if other_dd.type != 'USER']

    if isinstance(dd.children, list) and len(dd.children) > 0:
        domo_page.children = [
            cls._from_bootstrap(page_obj=child_dd, auth=auth)
            for child_dd in dd.children
            if child_dd.type == "page"
        ]

        [print(other_dd) for other_dd in dd.children
            if other_dd.type != "page"]

    return domo_page




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/DictDot.ipynb.

# %% auto 0
__all__ = ['DictDot']

# %% ../../nbs/utils/DictDot.ipynb 2
from types import SimpleNamespace

# %% ../../nbs/utils/DictDot.ipynb 3
class DictDot(SimpleNamespace):
    """a utility class that converts json into an object that supports data retrieval with dot notation"""

    def __init__(self, dictionary: dict,  # dictionary to convert
                 **kwargs):
        super().__init__(**kwargs)

        for key, value in dictionary.items():
            if isinstance(value, dict):
                self.__setattr__(key, DictDot(value))
            elif isinstance(value, list):
                new_list = []
                for item in value:
                    if isinstance(item, dict):
                        new_list.append(DictDot(item))
                    else:
                        new_list.append(item)
                self.__setattr__(key, new_list)
            else:
                self.__setattr__(key, value)

    def __getattr__(self, item):
        return None




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/client/95_DomoAuth.ipynb.

# %% auto 0
__all__ = ['get_full_auth', 'get_developer_auth', 'test_access_token', 'DomoAuth', 'InvalidCredentialsError',
           'InvalidInstanceError', 'DomoFullAuth', 'DomoTokenAuth', 'DomoDeveloperAuth']

# %% ../../nbs/client/95_DomoAuth.ipynb 3
from dataclasses import dataclass, field
from abc import abstractmethod
from typing import Optional, Union

import aiohttp

import domolibrary.client.ResponseGetData as rgd

# %% ../../nbs/client/95_DomoAuth.ipynb 5
async def get_full_auth(
    domo_instance: str,  # domo_instance.domo.com
    domo_username: str,  # email address
    domo_password: str,
    session: Optional[aiohttp.ClientSession] = None,
    debug_api: bool = False
) -> rgd.ResponseGetData:
    """uses username and password authentication to retrieve a full_auth access token"""

    is_close_session = False

    if not session:
        is_close_session = True
        session = aiohttp.ClientSession()

    url = f"https://{domo_instance}.domo.com/api/content/v2/authentication"

    tokenHeaders = {"Content-Type": "application/json"}

    body = {
        "method": "password",
        "emailAddress": domo_username,
        "password": domo_password,
    }

    if debug_api:
        print(body, url)

    res = await session.request(method="POST", url=url, headers=tokenHeaders, json=body)

    if is_close_session:
        await session.close()

    return await rgd.ResponseGetData._from_aiohttp_response(res)

# %% ../../nbs/client/95_DomoAuth.ipynb 12
async def get_developer_auth(
    domo_client_id: str,
    domo_client_secret: str,
    session: Optional[aiohttp.ClientSession] = None,
    debug_api: bool = False
) -> rgd.ResponseGetData:

    """
    only use for authenticating against apis documented under developer.domo.com
    """
    is_close_session = False

    if not session:
        is_close_session = True
        session = aiohttp.ClientSession(
            auth=aiohttp.BasicAuth(domo_client_id, domo_client_secret)
        )

    url = f"https://api.domo.com/oauth/token?grant_type=client_credentials"

    if debug_api:
        print(url, domo_client_id, domo_client_secret)

    res = await session.request(method="GET", url=url)

    if is_close_session:
        await session.close()

    return await rgd.ResponseGetData._from_aiohttp_response(res)

# %% ../../nbs/client/95_DomoAuth.ipynb 16
async def test_access_token(
    domo_access_token: str,  # as provided in Domo > Admin > Authentication > AccessTokens
    domo_instance: str,  # <domo_instance>.domo.com
    session: Optional[aiohttp.ClientSession] = None,
    debug_api: bool = False
):
    """
    will attempt to validate against the 'me' API.
    This is the same authentication test the Domo Java CLI uses.
    """

    is_close_session = False

    if not session:
        is_close_session = True
        session = aiohttp.ClientSession()

    url = f"https://{domo_instance}.domo.com/api/content/v2/users/me"

    tokenHeaders = {"X-DOMO-Developer-Token": domo_access_token}

    if debug_api:
        print(url,tokenHeaders)

    res = await session.request(method="GET", headers=tokenHeaders, url=url)

    if is_close_session:
        await session.close()

    return await rgd.ResponseGetData._from_aiohttp_response(res)

# %% ../../nbs/client/95_DomoAuth.ipynb 20
@dataclass
class _DomoAuth_Required:
    """required parameters for all Domo Auth classes"""

    domo_instance: str

    def __post_init__(self):
        if self.domo_instance:
            self.set_manual_login()

    def set_manual_login(self):
        self.url_manual_login = (
            f"https://{self.domo_instance}.domo.com/auth/index?domoManualLogin=true"
        )


@dataclass
class _DomoAuth_Optional:
    """parameters are defined after initialization"""

    token: Optional[str] = field(default=None, repr=False)
    token_name: Optional[str] = field(default=None)
    user_id: Optional[str] = field(default=None, repr=False)
    auth_header: dict = field(default_factory=dict, repr=False)

    url_manual_login: Optional[str] = None

    async def get_auth_token(self) -> Union[str, None]:
        """placeholder method"""
        pass

    async def generate_auth_header(self) -> Union[dict, None]:
        """returns auth header appropriate for this authentication method"""
        pass

    async def print_is_token(self, token_name=None) -> None:
        self.token_name = token_name or self.token_name

        if not self.token:
            await self.get_auth_token()

        token_str = f"{self.token_name} "

        if not self.token:
            print(
                f"🚧 failed to retrieve {token_str if token_name else ''}token from {self.domo_instance}")
            return False

        print(
            f"🎉 {token_str if token_name else ''}token retrieved from {self.domo_instance} ⚙️")
        return True


# %% ../../nbs/client/95_DomoAuth.ipynb 21
@dataclass
class DomoAuth(_DomoAuth_Optional, _DomoAuth_Required):
    """abstract DomoAuth class"""

# %% ../../nbs/client/95_DomoAuth.ipynb 25
class DomoErrror(Exception):
    """base exception"""

    def __init__(
        self,
        status: Optional[int] = None,  # API request status
        message: str = "error",  # <domo_instance>.domo.com
        domo_instance: Optional[str] = None,
    ):

        instance_str = f" at {domo_instance}" if domo_instance else ""
        status_str = f"Status {status} - " if status else ""
        self.message = f"{status_str}{message}{instance_str}"
        super().__init__(self.message)

# %% ../../nbs/client/95_DomoAuth.ipynb 26
class InvalidCredentialsError(DomoErrror):
    """return invalid credentials sent to API"""

    def __init__(
        self,
        status: Optional[int] = None,  # API request status
        message="invalid credentials",
        domo_instance: Optional[str] = None,
    ):

        super().__init__(status=status, message=message, domo_instance=domo_instance)


class InvalidInstanceError(DomoErrror):
    """return if invalid domo_instance sent to API"""

    def __init__(
        self,
        status: Optional[int] = None,
        message="invalid instance",
        domo_instance: Optional[str] = None,
    ):
        super().__init__(status=status, message=message, domo_instance=domo_instance)

# %% ../../nbs/client/95_DomoAuth.ipynb 29
@dataclass
class _DomoFullAuth_Required(_DomoAuth_Required):
    """mix requied parameters for DomoFullAuth"""

    domo_username: str
    domo_password: str = field(repr=False)


# %% ../../nbs/client/95_DomoAuth.ipynb 30
@dataclass
class DomoFullAuth(_DomoAuth_Optional, _DomoFullAuth_Required):
    """use for full authentication token"""

    def generate_auth_header(self, token: str) -> dict:
        self.auth_header = {"x-domo-authentication": token}
        return self.auth_header

    async def get_auth_token(
        self,
        session: Optional[aiohttp.ClientSession] = None,
        debug_api : bool = False
    ) -> str:
        """returns `token` if valid credentials provided else raises Exception and returns None"""

        res = await get_full_auth(
            domo_instance=self.domo_instance,
            domo_username=self.domo_username,
            domo_password=self.domo_password,
            session=session,
            debug_api = debug_api
        )

        if res.is_success and res.response.get("reason") == "INVALID_CREDENTIALS":
            raise InvalidCredentialsError(
                status=res.status,
                message=str(res.response.get("reason")),
                domo_instance=self.domo_instance,
            )

        if res.status == 403:
            raise InvalidInstanceError(
                status=res.status,
                message="INVALID INSTANCE",
                domo_instance=self.domo_instance,
            )

        token = str(res.response.get("sessionToken"))
        self.token = token
        self.user_id = str(res.response.get("userId"))

        self.auth_header = self.generate_auth_header(token=token)

        if not self.token_name:
            self.token_name = "full_auth"

        return self.token

# %% ../../nbs/client/95_DomoAuth.ipynb 36
@dataclass
class _DomoTokenAuth_Required(_DomoAuth_Required):
    """mix requied parameters for DomoFullAuth"""

    domo_access_token: str = field(repr=False)

# %% ../../nbs/client/95_DomoAuth.ipynb 37
@dataclass
class DomoTokenAuth(_DomoAuth_Optional, _DomoTokenAuth_Required):
    """
    use for access_token authentication.
    Tokens are generated in domo > admin > access token
    Necessary in cases where direct sign on is not permitted
    """

    def generate_auth_header(self, token: str) -> dict:
        self.auth_header = {"x-domo-developer-token": token}
        return self.auth_header

    async def get_auth_token(
        self, session: Optional[aiohttp.ClientSession] = None,
        debug_api : bool = False
    ) -> str:
        """
        updates internal attributes
        having an access_token assumes pre-authenticaiton
        """

        res = await test_access_token(
            domo_instance=self.domo_instance,
            domo_access_token=self.domo_access_token,
            session=session,
            debug_api = debug_api
        )

        if res.status == 401 and res.response == "Unauthorized":
            raise InvalidCredentialsError(
                status=res.status,
                message=res.response,
                domo_instance=self.domo_instance,
            )

        self.token = self.domo_access_token
        self.user_id = res.response.get("id")

        self.auth_header = self.generate_auth_header(token=self.token)

        if not self.token_name:
            self.token_name = "token_auth"

        return self.token

# %% ../../nbs/client/95_DomoAuth.ipynb 41
@dataclass
class _DomoDeveloperAuth_Required(_DomoAuth_Required):
    """mix requied parameters for DomoFullAuth"""

    domo_client_id: str
    domo_client_secret: str = field(repr=False)

# %% ../../nbs/client/95_DomoAuth.ipynb 42
@dataclass(init=False)
class DomoDeveloperAuth(_DomoAuth_Optional, _DomoDeveloperAuth_Required):
    """use for full authentication token"""

    def __init__(self, domo_client_id: str, domo_client_secret: str):
        self.domo_client_id = domo_client_id
        self.domo_client_secret = domo_client_secret
        self.domo_instance = ""

    def generate_auth_header(self, token: str) -> dict:
        self.auth_header = {"Authorization": "bearer " + token}
        return self.auth_header

    async def get_auth_token(
        self,
        session: Optional[aiohttp.ClientSession] = None,
        debug_api : bool = False
    ) -> str:

        res = await get_developer_auth(
            domo_client_id=self.domo_client_id,
            domo_client_secret=self.domo_client_secret,
            session=session,
            debug_api = debug_api
        )

        if res.status == 401:
            raise InvalidCredentialsError(
                status=res.status,
                message=str(res.response),
                domo_instance=self.domo_instance,
            )

        token = str(res.response.get("access_token"))
        self.token = token
        self.user_id = res.response.get("userId")
        self.domo_instance = res.response.get("domain")
        self.set_manual_login()

        self.auth_header = self.generate_auth_header(token=token)

        if not self.token_name:
            self.token_name = "developer_auth"

        return token



from ..DomoClasses import DomoAuth as dmda
from ..DomoClasses import DomoDataset as dmds
from .read_creds_from_dotenv import read_creds_from_dotenv, split_creds


def get_creds(array_env,
              array_key_starts_with,
              array_split,
              consol_env,
              consol_ds_id,
              debug: bool = False):
    instance_list = split_creds(env=array_env,
                                key_starts_with=array_key_starts_with,
                                env_var_list=array_split)

    if debug:
        print('\nconsol_env')
        print(consol_env, consol_env.CONSOL_INSTANCE)
        print('\ninstances to sync')
        for instance in instance_list:
            print(instance.domo_instance)

    consol_auth = dmda.DomoFullAuth(domo_instance=consol_env.CONSOL_INSTANCE,
                                    domo_username=consol_env.CONSOL_USERNAME,
                                    domo_password=consol_env.CONSOL_PASSWORD)

    consol_dataset = dmds.DomoDataset(full_auth=consol_auth, id=consol_ds_id)

    return instance_list, consol_auth, consol_dataset



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/client/99_ResponseGetData.ipynb.

# %% auto 0
__all__ = ['API_Response', 'STREAM_FILE_PATH', 'ResponseGetData']

# %% ../../nbs/client/99_ResponseGetData.ipynb 2
# pylint: disable=no-member
from dataclasses import dataclass, field
from typing import Optional

import json
import orjson


import asyncio
import requests
import aiohttp

from fastcore.utils import patch_to, patch

# %% ../../nbs/client/99_ResponseGetData.ipynb 4
API_Response = any


@dataclass
class ResponseGetData:
    """preferred response class for all API Requests"""

    status: int
    response: API_Response
    is_success: bool
    auth: dict = field(repr = False, default=None)

    def set_response(self, response):
        self.response = response

# %% ../../nbs/client/99_ResponseGetData.ipynb 8
@patch_to(ResponseGetData, cls_method=True)
def _from_requests_response(
    cls, res: requests.Response  # requests response object
) -> ResponseGetData:
    """returns ResponseGetData"""

    # JSON responses
    if res.ok and "application/json" in res.headers.get("Content-Type", {}):
        return cls(status=res.status_code, response=res.json(), is_success=True)

    # default text responses
    elif res.ok:
        return cls(status=res.status_code, response=res.text, is_success=True)

    # errors
    return cls(status=res.status_code, response=res.reason, is_success=False)

# %% ../../nbs/client/99_ResponseGetData.ipynb 12
@patch_to(ResponseGetData, cls_method=True)
def _from_httpx_response(
    cls, res: requests.Response,  # requests response object
    auth : Optional[any] = None,
    debug_api : bool = False
) -> ResponseGetData:
    """returns ResponseGetData"""

    # JSON responses
    ok = True if res.status_code <= 299 and res.status_code >= 200 else False
    
    if ok and "application/json" in res.headers.get("Content-Type", {}):
        return cls(status=res.status_code, response=res.json(), is_success=True, auth = auth)

    # default text responses
    elif ok:
        return cls(status=res.status_code, response=res.text, is_success=True, auth = auth)

    # errors
    return cls(status=res.status_code, response=res.reason_phrase, is_success=False, auth=auth)


# %% ../../nbs/client/99_ResponseGetData.ipynb 15
STREAM_FILE_PATH = '__large-file.json'

async def _write_stream(res: aiohttp.ClientResponse,
                        file_name: str = STREAM_FILE_PATH,
                        stream_chunks=10):
    
    print(type(res), type(res.content), stream_chunks)

    index = 0
    with open(file_name, 'wb') as fd:
        async for chunk in res.content.iter_chunked(1024):
            index +=1
            print(f"writing chunk - {index}")
            fd.write(chunk)

            print(res.content.at_eof())
        
    print('done writing stream')
    
    return None


async def _read_stream(file_name : str = STREAM_FILE_PATH):
    with open(STREAM_FILE_PATH, "rb") as f:
        return f.read()


# %% ../../nbs/client/99_ResponseGetData.ipynb 16
@patch(cls_method=True)
async def _from_aiohttp_response(
    cls: ResponseGetData, 
    res: aiohttp.ClientResponse,  # requests response object
    auth : Optional[any] = None,
    process_stream: bool = False,
    stream_chunks : int = 10,
    debug_api : bool = False
) -> ResponseGetData:

    """async method returns ResponseGetData"""
    if debug_api:
        print( f"ResponseGetData: res.ok = {res.ok} , res.status = {res.status}" )
    

    try:
        data = None

        if process_stream:
            await _write_stream(res = res, stream_chunks = stream_chunks)
            data =await _read_stream()
        
        else:        
            data = await res.text()
    
        if debug_api:
            print('converting to text complete')
    
    except asyncio.TimeoutError as e:
        print(f"ResponseGetDataError: {str(e)} , trying content.read")

        data = await res.content.read()

    if res.ok and "application/json" in res.headers.get("Content-Type", {}):
        return cls(status=res.status, response= orjson.loads(data), is_success=True, auth = auth)

    elif res.ok:
        return cls(status=res.status, response= data, is_success=True, auth = auth)

    # response is error
    else:
        return cls(status=res.status, response=res.reason, is_success=False, auth = auth)

# %% ../../nbs/client/99_ResponseGetData.ipynb 20
@patch(cls_method=True)
async def _from_looper(cls: ResponseGetData,
                       res: ResponseGetData,  # requests response object
                       array: list
                       ) -> ResponseGetData:

    """async method returns ResponseGetData"""

    if res.is_success:
        res.response = array
        return res

    # response is error
    else:
        return res




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoAccount.ipynb.

# %% auto 0
__all__ = ['DomoAccount_Config', 'DomoAccount_Config_AbstractCredential', 'DomoAccount_Config_DatasetCopy',
           'DomoAccount_Config_Governance', 'DomoAccount_Config_HighBandwidthConnector', 'AccountConfig', 'DomoAccount',
           'DomoAccount_UpdateName_Error', 'DomoAccount_CreateAccount_Error', 'DomoAccount_DeleteAccount_Error']

# %% ../../nbs/classes/50_DomoAccount.ipynb 3
from enum import Enum
from dataclasses import dataclass, field
from abc import ABC, abstractclassmethod, abstractmethod

import datetime as dt
import re


import aiohttp

from fastcore.basics import patch_to

import domolibrary.utils.convert as cd
import domolibrary.utils.DictDot as util_dd
import domolibrary.client.DomoAuth as dmda
import domolibrary.routes.account as account_routes

# %% ../../nbs/classes/50_DomoAccount.ipynb 5
class DomoAccount_Config(ABC):
    """DomoAccount Config abstract base class"""
    data_provider_type : str

    @abstractclassmethod
    def _from_json(cls, obj):
        """convert accounts API response into a class object"""
        pass

    @abstractmethod
    def to_json(self):
        """convert class object into a format the accounts API expects"""
        pass

# %% ../../nbs/classes/50_DomoAccount.ipynb 6
@dataclass
class DomoAccount_Config_AbstractCredential(DomoAccount_Config):
    data_provider_type = 'abstract-credential-store'
    credentials: dict

    def __post_init_(self):
        super().__init__(self)

    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(
            credentials=dd.credentials,
        )

    def to_json(self):
        return {"credentials": self.credentials}

# %% ../../nbs/classes/50_DomoAccount.ipynb 7
@dataclass
class DomoAccount_Config_DatasetCopy(DomoAccount_Config):
    domo_instance: str
    access_token: str = field(repr=False)

    data_provider_type = 'dataset-copy'
    

    def __post_init_(self):
        super().__init__(self)

    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(
            access_token=dd.accessToken,
            domo_instance=dd.instance
        )

    def to_json(self):
        return {"accessToken": self.access_token,
                "instance": self.domo_instance
                }


# %% ../../nbs/classes/50_DomoAccount.ipynb 8
@dataclass
class DomoAccount_Config_Governance(DomoAccount_Config):
    domo_instance: str
    access_token: str = field(repr=False)

    data_provider_type = 'domo-governance-d14c2fef-49a8-4898-8ddd-f64998005600'

    def __post_init_(self):
        super().__init__(self)

    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(
            access_token=dd.apikey,
            domo_instance=dd.customer
        )

    def to_json(self):
        return {"apikey": self.access_token,
                "customer": self.domo_instance
                }


# %% ../../nbs/classes/50_DomoAccount.ipynb 9
@dataclass
class DomoAccount_Config_HighBandwidthConnector(DomoAccount_Config):
    aws_access_key: str
    aws_secret_key: str = field(repr=False)
    s3_staging_dir: str
    region: str = "us-west-2"
    
    data_provider_type = "amazon-athena-high-bandwidth"

    def __post_init_(self):
        super().__init__(self)

    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(
            aws_access_key=dd.awsAccessKey,
            aws_secret_key=dd.awsSecretKey,
            s3_staging_dir=dd.s3StagingDir,
            region=dd.region,
        )

    def to_json(self):
        return {
            "awsAccessKey": self.aws_access_key,
            "awsSecretKey": self.aws_secret_key,
            "s3StagingDir": self.s3_staging_dir,
            "region": self.region,
        }

# %% ../../nbs/classes/50_DomoAccount.ipynb 10
class AccountConfig(Enum):
    """
    Enum provides appropriate spelling for data_provider_type and config object.  
    The name of the enum should correspond with the data_provider_type with hyphens replaced with underscores.
    """

    amazon_athena_high_bandwidth = DomoAccount_Config_HighBandwidthConnector

    abstract_credential_store = DomoAccount_Config_AbstractCredential

    dataset_copy = DomoAccount_Config_DatasetCopy

    domo_governance_d14c2fef_49a8_4898_8ddd_f64998005600 = DomoAccount_Config_Governance


# %% ../../nbs/classes/50_DomoAccount.ipynb 12
@dataclass
class DomoAccount:
    name: str
    data_provider_type: str

    id: int = None
    created_dt: dt.datetime = None
    modified_dt: dt.datetime = None
    auth: dmda.DomoAuth = field(repr=False, default=None)

    config: DomoAccount_Config = None

    @classmethod
    def _from_json(cls, obj: dict, auth: dmda.DomoAuth = None):
        """converts data_v1_accounts API response into an accounts class object"""

        dd = util_dd.DictDot(obj)

        return cls(
            id=dd.id,
            name=dd.displayName,
            data_provider_type=dd.dataProviderType,
            created_dt=cd.convert_epoch_millisecond_to_datetime(dd.createdAt),
            modified_dt=cd.convert_epoch_millisecond_to_datetime(dd.modifiedAt),
            auth=auth,
        )
    
    async def _get_config(self, session = None, debug_api : bool = None, return_raw: bool = False):
    
        res_config = await account_routes.get_account_config(
            auth=self.auth,
            account_id=self.id,
            data_provider_type=self.data_provider_type,
            session=session,
            debug_api=debug_api
        )

        if return_raw:
            return res_config

        enum_clean = re.sub("-", "_", self.data_provider_type)

        if not enum_clean in AccountConfig.__members__:
            return None

        self.config = (
            AccountConfig[enum_clean].value._from_json(res_config.response)
        )

        return self.config


# %% ../../nbs/classes/50_DomoAccount.ipynb 13
@patch_to(DomoAccount, cls_method=True)
async def get_from_id(
    cls,
    auth: dmda.DomoAuth,
    account_id: int,
    session: aiohttp.ClientSession = None,
    return_raw: bool = False,
    debug_api: bool = False
):

    res = await account_routes.get_account_from_id(
        auth=auth, account_id=account_id, session=session, debug_api = debug_api
    )
    if return_raw:
        return res

    if not res.is_success:
        return None

    obj = res.response
    acc = cls._from_json(obj, auth)

    await acc._get_config(session = session, debug_api = debug_api)

    return acc

# %% ../../nbs/classes/50_DomoAccount.ipynb 17
@patch_to(DomoAccount)
async def update_config(self: DomoAccount,
                        auth: dmda.DomoAuth = None,
                        debug_api: bool = False,
                        config : DomoAccount_Config = None,
                        session: aiohttp.ClientSession = None, return_raw: bool = False):

    auth = auth or self.auth

    config = config or self.config

    res = await account_routes.update_account_config(auth=auth,
                                                     account_id=self.id,
                                                     data_provider_type=self.data_provider_type,
                                                     config_body=config.to_json(),
                                                     debug_api=debug_api, session=session)

    if return_raw:
        return res

    await self._get_config(session=session, debug_api=debug_api)

    return self


# %% ../../nbs/classes/50_DomoAccount.ipynb 20
class DomoAccount_UpdateName_Error(Exception):
    pass

@patch_to(DomoAccount)
async def update_name(self: DomoAccount,
                      account_name: str = None,
                      auth: dmda.DomoAuth = None,
                      debug_api: bool = False,
                      session: aiohttp.ClientSession = None,
                      return_raw: bool = False):

    auth = auth or self.auth

    # print(auth, self.id, self.data_provider_type, self.config.to_json())

    res = await account_routes.update_account_name(auth=auth,
                                                   account_id=self.id,
                                                   account_name=account_name or self.name,
                                                   debug_api=debug_api,
                                                   session=session)

    if return_raw:
        return res

    if not res.is_success:
        raise DomoAccount_UpdateName_Error()
    
    self = await self.get_from_id(auth=auth, account_id=self.id)
    
    return self


# %% ../../nbs/classes/50_DomoAccount.ipynb 24
class DomoAccount_CreateAccount_Error(Exception):
    def __init__(self, account_name, domo_instance, status, reason):
        message = f"CreateAccount Error :: {account_name} in {domo_instance} :: {status} - {reason}"
        super.__init__(message)


@patch_to(DomoAccount, cls_method=True)
def generate_create_body(cls, account_name, config):
    return {'displayName': account_name,
            'dataProviderType': config.data_provider_type,
            'name': config.data_provider_type,
            'configurations': config.to_json()}


@patch_to(DomoAccount, cls_method=True)
async def create_account(cls: DomoAccount,
                         account_name: str, 
                         config: DomoAccount_Config,
                         auth: dmda.DomoAuth,
                         debug_api: bool = False,
                         session: aiohttp.ClientSession = None):

    body = cls.generate_create_body(account_name=account_name, config=config)

    res = await account_routes.create_account(auth=auth,
                                              config_body=body,
                                              debug_api=debug_api, session=session)

    if not res.is_success:
        raise DomoAccount_CreateAccount_Error(account_name=account_name,
                                              domo_instance=auth.domo_instance,
                                              status=res.status, reason=res.response)

    return await cls.get_from_id(auth=auth, account_id=res.response.get('id'))


# %% ../../nbs/classes/50_DomoAccount.ipynb 25
class DomoAccount_DeleteAccount_Error(Exception):
    def __init__(self, account_id, domo_instance, status, reason) -> bool:
        message = f"DeleteAccount Error :: {account_id} in {domo_instance} :: {status} - {reason}"
        super.__init__(message)


@patch_to(DomoAccount)
async def delete_account(self: DomoAccount,
                         auth: dmda.DomoAuth = None,
                         debug_api: bool = False, session: aiohttp.ClientSession = None):

    auth = auth or self.auth

    res = await account_routes.delete_account(auth=auth,
                                              account_id=self.id,
                                              debug_api=debug_api,
                                              session=session)

    if not res.is_success:
        raise DomoAccount_DeleteAccount_Error(
            account_id = self.id, domo_instance = auth.domo_instance, status = res.status, reason = res.response)

    return True




import aiohttp

from ...DomoClasses import DomoAuth as dmda
from ...utils.ResponseGetData import ResponseGetData
from . import get_data as gd


async def create_document(full_auth: dmda.DomoFullAuth, app_id: str,
                          domo_environment: str,
                          collection_name: str,
                          document: dict,
                          session: aiohttp.ClientSession = None,
                          debug: bool = False):
    url = f'https://{app_id}.domoapps.{domo_environment}.domo.com/domo/datastores/v1/collections/{collection_name}/documents'

    if debug:
        print(url)

    res = await gd.get_data(auth=full_auth,
                            method='POST',
                            url=url,
                            body=document,
                            session=session,
                            debug=debug)
    return res


async def get_documents(full_auth: dmda.DomoFullAuth, app_id: str,
                        domo_environment: str,
                        collection_name: str):
    url = f'https://{app_id}.domoapps.{domo_environment}.domo.com/domo/datastores/v1/collections/{collection_name}/documents/'

    res = await gd.get_data(auth=full_auth,
                            method='GET',
                            url=url)
    return res



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/integrations/DomoJupyter.ipynb.

# %% auto 0
__all__ = ['NoConfigCompanyError', 'GetInstanceConfig', 'GetDomains_Query_Exception_PW_Col_Error',
           'GetDomains_Query_AuthMatch_Error', 'GetJupyter_ErrorRetrievingAccount',
           'GetJupyter_ErrorRetrievingAccountProperty', 'get_jupyter_account', 'InvalidAccountTypeError',
           'DojoAccount_InstanceAuth', 'InvalidAccountNameError', 'GenerateAuth_InvalidDomoInstanceList',
           'GenerateAuth_CredentialsNotProvided']

# %% ../../nbs/integrations/DomoJupyter.ipynb 2
from dataclasses import dataclass, field
from typing import Optional
from enum import Enum


import re
import pandas as pd
import importlib
import datetime as dt
import time

import json

# import domojupyter as dj

import domolibrary.client.DomoAuth as dmda
import domolibrary.client.Logger as lc
import domolibrary.classes.DomoDataset as dmds

from nbdev.showdoc import show_doc
from fastcore.basics import patch_to


# %% ../../nbs/integrations/DomoJupyter.ipynb 3
class NoConfigCompanyError(Exception):
    def __init__(self, sql, domo_instance):
        message = f'SQL "{sql}" returned no results in {domo_instance}'
        self.message = message
        super().__init__(self.message)


class GetInstanceConfig:
    config: pd.DataFrame = None
    logger: lc.Logger = None

    def __init__(self, logger: Optional[lc.Logger] = None):

        self.logger = logger or lc.Logger(app_name='GetInstanceConfig')

    async def _retrieve_company_ds(self,
                                   config_auth: dmda.DomoAuth,
                                   dataset_id: str,
                                   sql: str,
                                   debug_prn: bool = False,
                                   debug_api: bool = False,
                                   debug_log: bool = False,
            
                                   ) -> pd.DataFrame:

        ds = await dmds.DomoDataset.get_from_id(auth=config_auth,
                                                dataset_id=dataset_id, debug_api=debug_api)

        message = f"⚙️ START - Retrieving company list \n{ds.display_url()} using \n{sql}"

        if debug_prn:
            print(message)

        self.logger.log_info(message, debug_log=debug_log)

        df = await ds.query_dataset_private(auth=config_auth,
                                            dataset_id=dataset_id,
                                            sql=sql,
                                            debug_api=debug_api)
        if len(df.index) == 0:
            raise NoConfigCompanyError(
                sql, domo_instance=config_auth.domo_instance)

        self.config = df

        message = f"\n⚙️ SUCCESS 🎉 Retrieved company list \nThere are {len(df.index)} companies to update"

        if debug_prn:
            print(message)
        logger.log_info(message, debug_log=debug_log)

        return df


# %% ../../nbs/integrations/DomoJupyter.ipynb 6
class GetDomains_Query_Exception_PW_Col_Error(Exception):
    """raise if SQL query fails to return column named 'config_exception_pw'"""

    def __init__(self, sql :str = None, domo_instance : str = None, message : str = None):
        message = message or f"Query failed to return a column 'config_exception_pw' sql = {sql} in {domo_instance}"
        super().__init__(self, message)


@patch_to(GetInstanceConfig, cls_method=True)
async def get_domains_with_global_config_auth(cls: GetInstanceConfig,

                                              config_dataset_id: str,
                                              config_auth: dmda.DomoAuth, # which instance to retrieve configuration data from
                                              
                                              global_auth: dmda.DomoAuth, # global authentication credentials
                                              global_exception_auth: dmda.DomoAuth, # exception credentials (ex 24 char password)
                                              # must return a column named domo_instance, if there is an exception_auth, must return a column 'config_exception_pw'
                                              
                                              config_sql: str = "select domain as domo_instance, config_exception_pw from table",
                                              
                                              debug_api: bool = False,
                                              debug_log: bool = False,
                                              debug_prn: bool = False,
                                              
                                              logger: lc.Logger = None) -> pd.DataFrame:
    if not logger:
        logger = lc.Logger(app_name='get_domains_with_global_config_auth')

    gic = cls(logger = logger)

    df = await gic._retrieve_company_ds(config_auth=config_auth,
                                        dataset_id=config_dataset_id,
                                        sql=config_sql,
                                       debug_prn=debug_prn,
                                       debug_log=debug_log,
                                       debug_api=debug_api)
    
    if 'config_exception_pw' not in df.columns:
        message = f"Query failed to return a column 'config_exception_pw' sql = {config_sql} in {config_auth.domo_instance}"
        gic.logger.log_error(message)
        raise GetDomains_Query_Exception_PW_Col_Error(message= message)

    for index, instance in df.iterrows():
        creds = global_auth

        if instance['config_exception_pw'] == 1:
            creds = global_exception_auth

        creds.domo_instance=instance['domo_instance']

        try:
            await creds.get_auth_token()
            df.at[index, 'is_valid'] = 1

        except dmda.InvalidCredentialsError as e:
            if debug_prn:
                print(e)
            
            logger.log_error(str(e))
            df.at[index, 'is_valid'] = 0
        
        finally:
            df.at[index, 'instance_auth'] = creds

    return df


# %% ../../nbs/integrations/DomoJupyter.ipynb 10
class GetDomains_Query_AuthMatch_Error(Exception):
    """raise if SQL query fails to return column named 'auth_match_col'"""
    def __init__(self, sql : str = None, domo_instance : str = None ,message : str = None):
        self.message = message or f"Query failed to return a column 'auth_match_col' sql = {sql} in {domo_instance}"
        super().__init__( self.message)

@patch_to(GetInstanceConfig, cls_method=True)
async def get_domains_with_instance_auth(cls: GetInstanceConfig,
                                         config_dataset_id: str, # dataset_id to run config_sql query against
                                         config_auth: dmda.DomoAuth,  # which instance to retrieve configuration data from
                                         
                                         default_auth : dmda.DomoAuth, # default auth to use with each row
                                         auth_enum : Enum, # Enum where enum_name should match to `auth_match_col` from config_sql query and enum_value is the appropriate DomoAuth object
                                         
                                         config_sql: str = "select domain as domo_instance,concat(config_useprod, '-', project) as auth_match_col from table",
                                         
                                         debug_api: bool = False,
                                         debug_log: bool = False,
                                         debug_prn: bool = False,
                                         logger: lc.Logger = None # pass in Logger class
                                         ) -> pd.DataFrame: # returns a dataframe with domo_instance, instance_auth, and binary column is_valid 
    """uses a sql query to retrieve a list of domo_instances and map authentication object to each instance"""
    
    if not logger:
        logger = lc.Logger(app_name='get_domains_with_instance_auth')

    gic = cls(logger=logger)

    df = await gic._retrieve_company_ds(config_auth=config_auth,
                                        dataset_id=config_dataset_id,
                                        sql=config_sql,
                                        debug_prn=debug_prn,
                                        debug_log=debug_log,
                                        debug_api=debug_api)

    if 'auth_match_col' not in df.columns:
        message = f"Query failed to return a column 'auth_match_col' sql = {config_sql} in {config_auth.domo_instance}"
        raise GetDomains_Query_AuthMatch_Error(message)

    for index, instance in df.iterrows():
        
        creds = default_auth

        match_auth = next(
            (member.value for member in auth_enum if member.name == instance['auth_match_col']))

        if match_auth:
            creds = match_auth

        creds.domo_instance=instance['domo_instance']

        try:
            await creds.get_auth_token(debug_api = debug_api)
            df.at[index, 'is_valid'] = 1

        except dmda.InvalidCredentialsError as e:
            if debug_prn:
                print(e)
            
            logger.log_error(str(e))
            df.at[index, 'is_valid'] = 0
        
        finally:
            df.at[index, 'instance_auth'] = creds

    return df


# %% ../../nbs/integrations/DomoJupyter.ipynb 14
class GetJupyter_ErrorRetrievingAccount(Exception):
    def __init__(self, account_name):
        self.message = f'failure to retrieve DomoDojo Account {account_name}'
        super().__init__(self.message)


class GetJupyter_ErrorRetrievingAccountProperty(Exception):
    def __init__(self, account_name, property_name):
        self.message = f'failure to retrieve {property_name} DomoDojo Account {account_name}'
        super().__init__(self.message)

def get_jupyter_account(account_name: str,  # name of account as it appears in the
                        domojupyter_fn : callable,
                        maximum_retry: int = 10
                        ) -> (list, dict):  # returns account properties list and a dictionary of the properties.
    """import a domojupyter account, will loop until success"""
    account_properties = None

    retry_attempt = 0
    while not account_properties and retry_attempt <= maximum_retry:
        try:
            account_properties = domojupyter_fn.get_account_property_keys(
                account_name)
            retry_attempt += 1

        except GetJupyter_ErrorRetrievingAccount as e:
            return None

        except Exception as e:
            print(f"trying again - {account_name}")
            time.sleep(2)
    
    if not account_properties:
        raise GetJupyter_ErrorRetrievingAccountProperty(
            account_name=account_name)

    obj = {}

    retry_attempt = 0
    for index, prop in enumerate(account_properties):
        value = None

        while not value and retry_attempt <= maximum_retry:
            try:
                value = domojupyter_fn.get_account_property_value(
                    account_name, account_properties[index])
            
            except Exception as e:
                print(f"trying again - {prop}")
                time.sleep(2)

        if not value:
            raise GetJupyter_ErrorRetrievingAccountProperty(account_name=account_name, property_name = prop)
            
        obj.update({prop: value})

    return account_properties, obj

# %% ../../nbs/integrations/DomoJupyter.ipynb 15
class InvalidAccountTypeError(Exception):
    """raised when account type is not expected type"""

    def __init__(self, account_name, account_type):

        self.message = f"account: {account_name} is not {account_type}"

        super().__init__(self.message)
    pass

@dataclass
class DojoAccount_InstanceAuth:

    account_name: str

    domo_username: str = None

    display_name: str = field(repr=False, default=None)

    domo_instance: str = field(repr=False, default=None)
    domo_instance_ls: list = field(repr=False, default=None)

    raw_cred: dict = field(repr=False, default=None)
    domo_password: str = field(repr=False, default=None)
    domo_access_token : str = field(repr = False, default = None)

    auth_ls: list = field(repr=False, default=None)

    account_name_mask = '^dj_.*_acc'

    def __post_init__(self):
        if not self.display_name and self.domo_username:
            self._set_display_name()

    def _set_display_name(self ):
        clean_text = re.sub('@.*$', '', self.domo_username)
        self.display_name = clean_text


    @staticmethod
    def _test_regex_mask(test_string: str  # the string to test
                         , regex_mask: str  # the regex expression to test
                         ) -> bool:  # boolean of the re match
        """tests if a string matches the regex pattern"""

        return bool(re.match(regex_mask, test_string))

    @staticmethod
    def _clean_account_admin_accounts(account_name):

        clean_str = re.sub('^dj_', '', account_name)
        clean_str = re.sub('_acc$', '', clean_str)

        return clean_str

# %% ../../nbs/integrations/DomoJupyter.ipynb 16
class InvalidAccountNameError(Exception):
    """raised when account name does not follow format string"""

    def __init__(self, account_name=None, regex_pattern=None):
        account_str = f'"{account_name}" '
        regex_str = f'"{regex_pattern}"'

        message = f"string {account_str if account_name else ''}does not match regex pattern {regex_str or ''}"
        self.message = message

        super().__init__(self.message)

@patch_to(DojoAccount_InstanceAuth, cls_method=True)
def get_domo_instance_auth_account(cls: DojoAccount_InstanceAuth, account_name: str,  # domojupyter account to retrieve
                     # Domo's domojupyter module, pass in b/c can only be retrieved inside Domo jupyter notebook environment
                     domojupyter_fn: callable,
                     # set the domo_instance or retrieve from the domojupyter_account credential store
                     domo_instance=None
                     ):
    """
    retrieves Abstract Credential Store from DomoJupyter environment.
    expects credentials property to contain DOMO_USERNAME, DOMO_PASSWORD, or DOMO_ACCESS_TOKEN, and (optional) DOMO_INSTANCE
    """

    if not cls._test_regex_mask(account_name, cls.account_name_mask):
        raise InvalidAccountNameError(account_name = account_name, regex_pattern = cls.account_name_mask)

    account_properties, dj_account = get_jupyter_account(
        account_name, domojupyter_fn=domojupyter_fn)

    if account_properties != ['credentials']:
        raise InvalidAccountTypeError(
            account_name=account_name, account_type="abstract_credential_store")

    creds = json.loads(dj_account.get('credentials'))

    return cls(
        account_name=account_name,
        raw_cred=creds,
        domo_username=creds.get('DOMO_USERNAME'),
        domo_password=creds.get('DOMO_PASSWORD'),
        domo_access_token=creds.get('DOMO_ACCESS_TOKEN'),
        domo_instance=domo_instance or creds.get('DOMO_INSTANCE')
    )


# %% ../../nbs/integrations/DomoJupyter.ipynb 18
class GenerateAuth_InvalidDomoInstanceList(Exception):
    def __init__(self):
        message = f"provide a list of domo_instances"
        super().__init__(message)


class GenerateAuth_CredentialsNotProvided(Exception):
    def __init__(self):
        message = f"object does not have a valid combination of credentials (access_token or username and password)"
        super().__init__(message)


@patch_to(DojoAccount_InstanceAuth)
def _generate_auth(self, domo_instance):
    if self.domo_access_token:
            auth = dmda.DomoTokenAuth(
                domo_instance=domo_instance, domo_access_token=self.domo_access_token)

    elif self.domo_username and self.domo_password:

        auth = dmda.DomoFullAuth(domo_instance=domo_instance,
                                    domo_username=self.domo_username,
                                    domo_password=self.domo_password)
    
    else:
            raise GenerateAuth_CredentialsNotProvided()
    
    return auth

@patch_to(DojoAccount_InstanceAuth)
def generate_auth_ls(self : DojoAccount_InstanceAuth,
                          domo_instance_ls: list[str] = None # list of domo_instances
                          ) -> list[dmda.DomoAuth]: # list of domo auth objects
                          
    """for every domo_instance in domo_instance_ls generates an DomoAuth object"""

    # reset internal lists
    self.domo_instance = None

    self.domo_instance_ls = list(
        set(domo_instance_ls or self.domo_instance_ls))

    if not self.domo_instance_ls:
        raise GenerateAuth_InvalidDomoInstanceList()

    self.auth_ls = []
    for domo_instance in self.domo_instance_ls:
        auth = self._generate_auth(domo_instance)

        self.auth_ls.append(auth)

    return self.auth_ls



import aiohttp
import re
from dataclasses import field, dataclass
from sql_metadata import Parser
from typing import List, Any

from .DomoAuth import DomoFullAuth
import Library.DomoClasses.DomoDatacenter as dmdc
import Library.utils.LoggerClass as lc
from .routes import stream_routes
from ..utils.Base import Base
from ..utils.DictDot import DictDot


custom_query = ['enteredCustomQuery', 'query', 'customQuery']


@dataclass
class StreamConfig:
    name: str
    type: str
    value: str
    value_clean: str = None


@dataclass
class DomoStream(Base):
    id: str
    dataset_id: str
    transport_description: str
    transport_version: int
    update_method: str
    data_provider_name: str
    data_provider_key: str
    account_id: str = None
    account_display_name: str = None
    account_userid: str = None

    configuration: list[StreamConfig] = field(default_factory=list)
    configuration_tables: list[str] = field(default_factory=list)
    configuration_query: str = None

    @classmethod
    async def get_definition(cls,
                             full_auth: DomoFullAuth, 
                             stream_id: str, 
                             session: aiohttp.ClientSession = None, 
                             logger: lc.MyLogger = None):
        if logger : 
            logger.log_info (f"⚙️ get_definition: for {stream_id}")
        if stream_id is None:
            return None

        res = await stream_routes.get_stream_definition(full_auth=full_auth,
                                                        stream_id=stream_id,
                                                        session=session
                                                        )

        if res.status != 200:
            error_str = f"get_definition: error retrieving stream {stream_id} from {full_auth.domo_instance}"
            print(error_str)
            if logger : 
                logger.log_error(error_str)
            return None

        dd = DictDot(res.response)

        sd = cls(
            id=dd.id,
            transport_description=dd.transport.description,
            transport_version=dd.transport.version,
            update_method=dd.updateMethod,
            data_provider_name=dd.dataProvider.name,
            data_provider_key=dd.dataProvider.key,
            dataset_id=dd.dataSource.id
        )
        if dd.account:
            sd.account_id = dd.account.id
            sd.account_display_name = dd.account.displayName
            sd.account_userid = dd.account.userId

        sd.configuration = []

        for config in dd.configuration:
            sc = StreamConfig(
                name=config.name,
                type=config.type,
                value=config.value)

            if sc.name in custom_query:
                sc.value_clean = sc.value.replace('\n', ' ')
                sc.value_clean = re.sub(' +', ' ', sc.value_clean)
                sd.configuration_query = sc.value_clean

                try:
                    for table in Parser(sc.value).tables:
                        sd.configuration_tables.append(table)
                    sd.configuration_tables = sorted(
                        list(set(sd.configuration_tables)))
                    
                except Exception as e:
                    if logger : 
                        logger.log_error(f"get_definition: unable to parse table for stream {stream_id}. Exception : {e}")
                    print('ALERT: unable to parse table')
                    sd.configuration_tables = ['unable to auto-parse query']

            sd.configuration.append(sc)
        return sd

    @classmethod
    async def create_stream(cls,
                            cnfg_body,
                            full_auth: DomoFullAuth = None,
                            session: aiohttp.ClientSession = None,
                            debug: bool = False,
                            log_result: bool = False):
        return await stream_routes.create_stream(full_auth=full_auth,
                                                 body=cnfg_body,
                                                 session=session,
                                                 debug=debug)

    @classmethod
    async def update_stream(cls,
                            cnfg_body,
                            stream_id,
                            full_auth: DomoFullAuth = None,
                            session: aiohttp.ClientSession = None,
                            debug: bool = False,
                            log_result: bool = False):

        return await stream_routes.update_stream(full_auth=full_auth,
                                                 stream_id=stream_id,
                                                 body=cnfg_body,
                                                 session=session,
                                                 debug=debug)

    @classmethod
    async def upsert_connector(cls,
                               cnfg_body,
                               match_name=None,
                               full_auth: DomoFullAuth = None,
                               session: aiohttp.ClientSession = None,
                               debug: bool = False,
                               log_result: bool = False):
        search_body = dmdc.DomoDatacenter.generate_search_datacenter_body_by_name(
            entity_name=match_name)

        search_res = await dmdc.DomoDatacenter.search_datacenter(full_auth=full_auth,
                                                                 body=search_body,
                                                                 session=session,
                                                                 debug=debug,
                                                                 log_result=log_result)

        existing_ds = next((ds for ds in search_res if ds.get(
            'name').lower() == match_name.lower()), None)

        if debug:
            print(
                f"existing_ds - {existing_ds.id if existing_ds else ' not found '}")

        if existing_ds:
            existing_ds = await DomoDataset.getDomoProps(id=existing_ds.get('databaseId'),
                                                         full_auth=full_auth)
            return await cls.update_stream(cnfg_body,
                                           stream_id=existing_ds.stream_id,
                                           full_auth=full_auth,
                                           session=session,
                                           debug=False,
                                           log_result=False)
        else:
            return await cls.create_stream(cnfg_body,
                                           full_auth=full_auth,
                                           session=session,
                                           debug=debug,
                                           log_result=log_result)



import aiohttp
from pprint import pprint

from .get_data import get_data, looper
from ..DomoAuth import DomoFullAuth
from ...utils.ResponseGetData import ResponseGetData


def generate_search_datacenter_body(entities_list: list[str] = ['DATASET'],
                                    filters: list[dict] = None,
                                    combineResults: bool = True,
                                    count: int = 10,
                                    offset: int = 0):
    return {
        "entities": entities_list,
        "filters": filters or [],
        "combineResults": combineResults,
        "query": "*",
        "count": count,
        "offset": offset}

def generate_search_datacenter_account_body( search_str : str, is_exact_match:bool = True):
    return {
        "count": 100,
        "offset": 0,
        "combineResults": False,
        "query": search_str if is_exact_match else f"*{search_str}*",
        "filters": [],
        "facetValuesToInclude": [
            "DATAPROVIDERNAME",
            "OWNED_BY_ID",
            "VALID",
            "USED",
            "LAST_MODIFIED_DATE"
        ],
        "queryProfile": "GLOBAL",
        "entityList": [
            [
                "account"
            ]
        ],
        "sort": {
            "fieldSorts": [
                {
                    "field": "display_name_sort",
                    "sortOrder": "ASC"
                }
            ]
        }
    }
        


async def search_datacenter(full_auth: DomoFullAuth,
                            arr_fn: callable,
                            alter_maximum_fn: callable,
                            maximum: int = None,
                            body: dict = None,
                            session: aiohttp.ClientSession = None,
                            limit=1000,
                            debug: bool = False, log_result: bool = False) -> ResponseGetData:
    is_close_session = False
    if not session:
        session = aiohttp.ClientSession()
        is_close_session = True

    if not body:
        body = {
            "entities": ["DATASET"],
            "filters": [],
            "combineResults": False,
            "query": "*"}

    url = f"https://{full_auth.domo_instance}.domo.com/api/search/v1/query"

    if debug:
        print(url)
        pprint(body)

    res = await looper(auth=full_auth,
                       session=session,
                       url=url,
                       body=body,
                       offset_params={
                           'offset': 'offset',
                           'limit': 'count'
                       },
                       arr_fn=arr_fn,
                       alter_maximum_fn=alter_maximum_fn,
                       method='POST',
                       offset_params_in_body=True,
                       limit=limit,
                       maximum=maximum,
                       debug=debug)

    if is_close_session:
        await session.close()

    return res


async def get_lineage_upstream(full_auth: DomoFullAuth,
                               entity_type: str,
                               entity_id: str,
                               session: aiohttp.ClientSession = None,
                               debug: bool = False):
    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v1/lineage/{entity_type}/{entity_id}"

    params = {'traverseDown': 'false'}

    return await get_data(
        auth=full_auth,
        method="GET",
        url=url,
        params=params,
        session=session,
        debug=debug
    )



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/domo_library',
                'doc_host': 'https://jaewilson07.github.io',
                'git_url': 'https://github.com/jaewilson07/domo_library',
                'lib_path': 'domolibrary'},
  'syms': { 'domolibrary.classes.DomoAccount': { 'domolibrary.classes.DomoAccount.AccountConfig': ( 'classes/domoaccount.html#accountconfig',
                                                                                                    'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount': ( 'classes/domoaccount.html#domoaccount',
                                                                                                  'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount._from_json': ( 'classes/domoaccount.html#domoaccount._from_json',
                                                                                                             'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount._get_config': ( 'classes/domoaccount.html#domoaccount._get_config',
                                                                                                              'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount.create_account': ( 'classes/domoaccount.html#domoaccount.create_account',
                                                                                                                 'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount.delete_account': ( 'classes/domoaccount.html#domoaccount.delete_account',
                                                                                                                 'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount.generate_create_body': ( 'classes/domoaccount.html#domoaccount.generate_create_body',
                                                                                                                       'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount.get_from_id': ( 'classes/domoaccount.html#domoaccount.get_from_id',
                                                                                                              'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount.update_config': ( 'classes/domoaccount.html#domoaccount.update_config',
                                                                                                                'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount.update_name': ( 'classes/domoaccount.html#domoaccount.update_name',
                                                                                                              'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config': ( 'classes/domoaccount.html#domoaccount_config',
                                                                                                         'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config._from_json': ( 'classes/domoaccount.html#domoaccount_config._from_json',
                                                                                                                    'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config.to_json': ( 'classes/domoaccount.html#domoaccount_config.to_json',
                                                                                                                 'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_AbstractCredential': ( 'classes/domoaccount.html#domoaccount_config_abstractcredential',
                                                                                                                            'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_AbstractCredential.__post_init_': ( 'classes/domoaccount.html#domoaccount_config_abstractcredential.__post_init_',
                                                                                                                                         'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_AbstractCredential._from_json': ( 'classes/domoaccount.html#domoaccount_config_abstractcredential._from_json',
                                                                                                                                       'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_AbstractCredential.to_json': ( 'classes/domoaccount.html#domoaccount_config_abstractcredential.to_json',
                                                                                                                                    'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_DatasetCopy': ( 'classes/domoaccount.html#domoaccount_config_datasetcopy',
                                                                                                                     'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_DatasetCopy.__post_init_': ( 'classes/domoaccount.html#domoaccount_config_datasetcopy.__post_init_',
                                                                                                                                  'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_DatasetCopy._from_json': ( 'classes/domoaccount.html#domoaccount_config_datasetcopy._from_json',
                                                                                                                                'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_DatasetCopy.to_json': ( 'classes/domoaccount.html#domoaccount_config_datasetcopy.to_json',
                                                                                                                             'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_Governance': ( 'classes/domoaccount.html#domoaccount_config_governance',
                                                                                                                    'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_Governance.__post_init_': ( 'classes/domoaccount.html#domoaccount_config_governance.__post_init_',
                                                                                                                                 'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_Governance._from_json': ( 'classes/domoaccount.html#domoaccount_config_governance._from_json',
                                                                                                                               'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_Governance.to_json': ( 'classes/domoaccount.html#domoaccount_config_governance.to_json',
                                                                                                                            'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_HighBandwidthConnector': ( 'classes/domoaccount.html#domoaccount_config_highbandwidthconnector',
                                                                                                                                'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_HighBandwidthConnector.__post_init_': ( 'classes/domoaccount.html#domoaccount_config_highbandwidthconnector.__post_init_',
                                                                                                                                             'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_HighBandwidthConnector._from_json': ( 'classes/domoaccount.html#domoaccount_config_highbandwidthconnector._from_json',
                                                                                                                                           'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_Config_HighBandwidthConnector.to_json': ( 'classes/domoaccount.html#domoaccount_config_highbandwidthconnector.to_json',
                                                                                                                                        'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_CreateAccount_Error': ( 'classes/domoaccount.html#domoaccount_createaccount_error',
                                                                                                                      'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_CreateAccount_Error.__init__': ( 'classes/domoaccount.html#domoaccount_createaccount_error.__init__',
                                                                                                                               'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_DeleteAccount_Error': ( 'classes/domoaccount.html#domoaccount_deleteaccount_error',
                                                                                                                      'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_DeleteAccount_Error.__init__': ( 'classes/domoaccount.html#domoaccount_deleteaccount_error.__init__',
                                                                                                                               'domolibrary/classes/DomoAccount.py'),
                                                 'domolibrary.classes.DomoAccount.DomoAccount_UpdateName_Error': ( 'classes/domoaccount.html#domoaccount_updatename_error',
                                                                                                                   'domolibrary/classes/DomoAccount.py')},
            'domolibrary.classes.DomoActivityLog': { 'domolibrary.classes.DomoActivityLog.ActivityLog_ObjectType': ( 'classes/domoactivitylog.html#activitylog_objecttype',
                                                                                                                     'domolibrary/classes/DomoActivityLog.py'),
                                                     'domolibrary.classes.DomoActivityLog.DomoActivityLog': ( 'classes/domoactivitylog.html#domoactivitylog',
                                                                                                              'domolibrary/classes/DomoActivityLog.py'),
                                                     'domolibrary.classes.DomoActivityLog.DomoActivityLog._process_activity_log_row': ( 'classes/domoactivitylog.html#domoactivitylog._process_activity_log_row',
                                                                                                                                        'domolibrary/classes/DomoActivityLog.py'),
                                                     'domolibrary.classes.DomoActivityLog.DomoActivityLog.get_activity_log': ( 'classes/domoactivitylog.html#domoactivitylog.get_activity_log',
                                                                                                                               'domolibrary/classes/DomoActivityLog.py')},
            'domolibrary.classes.DomoBootstrap': { 'domolibrary.classes.DomoBootstrap.DomoBootstrap': ( 'classes/.domobootstrap.html#domobootstrap',
                                                                                                        'domolibrary/classes/DomoBootstrap.py'),
                                                   'domolibrary.classes.DomoBootstrap.DomoBootstrap.get_all': ( 'classes/.domobootstrap.html#domobootstrap.get_all',
                                                                                                                'domolibrary/classes/DomoBootstrap.py'),
                                                   'domolibrary.classes.DomoBootstrap.DomoBootstrap.get_features': ( 'classes/.domobootstrap.html#domobootstrap.get_features',
                                                                                                                     'domolibrary/classes/DomoBootstrap.py'),
                                                   'domolibrary.classes.DomoBootstrap.DomoBootstrap.get_pages': ( 'classes/.domobootstrap.html#domobootstrap.get_pages',
                                                                                                                  'domolibrary/classes/DomoBootstrap.py'),
                                                   'domolibrary.classes.DomoBootstrap.DomoBootstrap_Feature': ( 'classes/.domobootstrap.html#domobootstrap_feature',
                                                                                                                'domolibrary/classes/DomoBootstrap.py'),
                                                   'domolibrary.classes.DomoBootstrap.DomoBootstrap_Feature._from_json_bootstrap': ( 'classes/.domobootstrap.html#domobootstrap_feature._from_json_bootstrap',
                                                                                                                                     'domolibrary/classes/DomoBootstrap.py')},
            'domolibrary.classes.DomoDataset': { 'domolibrary.classes.DomoDataset.DatasetSchema_AuthNotProvidedError': ( 'classes/domodataset.html#datasetschema_authnotprovidederror',
                                                                                                                         'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DatasetSchema_AuthNotProvidedError.__init__': ( 'classes/domodataset.html#datasetschema_authnotprovidederror.__init__',
                                                                                                                                  'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DatasetSchema_DatasetNotProvidedError': ( 'classes/domodataset.html#datasetschema_datasetnotprovidederror',
                                                                                                                            'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DatasetSchema_DatasetNotProvidedError.__init__': ( 'classes/domodataset.html#datasetschema_datasetnotprovidederror.__init__',
                                                                                                                                     'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DatasetSchema_Types': ( 'classes/domodataset.html#datasetschema_types',
                                                                                                          'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DatasetTags_AuthNotProvidedError': ( 'classes/domodataset.html#datasettags_authnotprovidederror',
                                                                                                                       'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DatasetTags_AuthNotProvidedError.__init__': ( 'classes/domodataset.html#datasettags_authnotprovidederror.__init__',
                                                                                                                                'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DatasetTags_SetTagsError': ( 'classes/domodataset.html#datasettags_settagserror',
                                                                                                               'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DatasetTags_SetTagsError.__init__': ( 'classes/domodataset.html#datasettags_settagserror.__init__',
                                                                                                                        'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset': ( 'classes/domodataset.html#domodataset',
                                                                                                  'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset.__post_init__': ( 'classes/domodataset.html#domodataset.__post_init__',
                                                                                                                'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset.create': ( 'classes/domodataset.html#domodataset.create',
                                                                                                         'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset.delete': ( 'classes/domodataset.html#domodataset.delete',
                                                                                                         'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset.display_url': ( 'classes/domodataset.html#domodataset.display_url',
                                                                                                              'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset.get_from_id': ( 'classes/domodataset.html#domodataset.get_from_id',
                                                                                                              'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset.index_dataset': ( 'classes/domodataset.html#domodataset.index_dataset',
                                                                                                                'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset.list_partitions': ( 'classes/domodataset.html#domodataset.list_partitions',
                                                                                                                  'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset.query_dataset_private': ( 'classes/domodataset.html#domodataset.query_dataset_private',
                                                                                                                        'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset.upload_data': ( 'classes/domodataset.html#domodataset.upload_data',
                                                                                                              'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_CreateDataset_Error': ( 'classes/domodataset.html#domodataset_createdataset_error',
                                                                                                                      'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_CreateDataset_Error.__init__': ( 'classes/domodataset.html#domodataset_createdataset_error.__init__',
                                                                                                                               'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_DeleteDataset_Error': ( 'classes/domodataset.html#domodataset_deletedataset_error',
                                                                                                                      'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_DeleteDataset_Error.__init__': ( 'classes/domodataset.html#domodataset_deletedataset_error.__init__',
                                                                                                                               'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_Schema': ( 'classes/domodataset.html#domodataset_schema',
                                                                                                         'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_Schema.get': ( 'classes/domodataset.html#domodataset_schema.get',
                                                                                                             'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_Schema_Column': ( 'classes/domodataset.html#domodataset_schema_column',
                                                                                                                'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_Schema_Column._from_json': ( 'classes/domodataset.html#domodataset_schema_column._from_json',
                                                                                                                           'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_Tags': ( 'classes/domodataset.html#domodataset_tags',
                                                                                                       'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_Tags._have_prereqs': ( 'classes/domodataset.html#domodataset_tags._have_prereqs',
                                                                                                                     'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_Tags.add': ( 'classes/domodataset.html#domodataset_tags.add',
                                                                                                           'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_Tags.get': ( 'classes/domodataset.html#domodataset_tags.get',
                                                                                                           'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_Tags.remove': ( 'classes/domodataset.html#domodataset_tags.remove',
                                                                                                              'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_Tags.set': ( 'classes/domodataset.html#domodataset_tags.set',
                                                                                                           'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_UploadData_DatasetUploadId_Error': ( 'classes/domodataset.html#domodataset_uploaddata_datasetuploadid_error',
                                                                                                                                   'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_UploadData_DatasetUploadId_Error.__init__': ( 'classes/domodataset.html#domodataset_uploaddata_datasetuploadid_error.__init__',
                                                                                                                                            'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_UploadData_Error': ( 'classes/domodataset.html#domodataset_uploaddata_error',
                                                                                                                   'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_UploadData_Error.__init__': ( 'classes/domodataset.html#domodataset_uploaddata_error.__init__',
                                                                                                                            'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_UploadData_UploadData_Error': ( 'classes/domodataset.html#domodataset_uploaddata_uploaddata_error',
                                                                                                                              'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_UploadData_UploadData_Error.DomoDataset_UploadData_CommitDatasetUploadId_Error': ( 'classes/domodataset.html#domodataset_uploaddata_uploaddata_error.domodataset_uploaddata_commitdatasetuploadid_error',
                                                                                                                                                                                 'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.DomoDataset_UploadData_UploadData_Error.__init__': ( 'classes/domodataset.html#domodataset_uploaddata_uploaddata_error.__init__',
                                                                                                                                       'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.QueryExecutionError': ( 'classes/domodataset.html#queryexecutionerror',
                                                                                                          'domolibrary/classes/DomoDataset.py'),
                                                 'domolibrary.classes.DomoDataset.QueryExecutionError.__init__': ( 'classes/domodataset.html#queryexecutionerror.__init__',
                                                                                                                   'domolibrary/classes/DomoDataset.py')},
            'domolibrary.classes.DomoPage': { 'domolibrary.classes.DomoPage.DomoPage': ( 'classes/domopage.html#domopage',
                                                                                         'domolibrary/classes/DomoPage.py'),
                                              'domolibrary.classes.DomoPage.DomoPage._from_bootstrap': ( 'classes/domopage.html#domopage._from_bootstrap',
                                                                                                         'domolibrary/classes/DomoPage.py'),
                                              'domolibrary.classes.DomoPage.DomoPage.display_url': ( 'classes/domopage.html#domopage.display_url',
                                                                                                     'domolibrary/classes/DomoPage.py')},
            'domolibrary.classes.DomoUser': { 'domolibrary.classes.DomoUser.DomoUser': ( 'classes/domouser.html#domouser',
                                                                                         'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUser._from_bootstrap_json': ( 'classes/domouser.html#domouser._from_bootstrap_json',
                                                                                                              'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUser._from_search_json': ( 'classes/domouser.html#domouser._from_search_json',
                                                                                                           'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUser._from_virtual_json': ( 'classes/domouser.html#domouser._from_virtual_json',
                                                                                                            'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUser.create_user': ( 'classes/domouser.html#domouser.create_user',
                                                                                                     'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUser.request_password_reset': ( 'classes/domouser.html#domouser.request_password_reset',
                                                                                                                'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUser.reset_password': ( 'classes/domouser.html#domouser.reset_password',
                                                                                                        'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUser.set_user_landing_page': ( 'classes/domouser.html#domouser.set_user_landing_page',
                                                                                                               'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUsers': ( 'classes/domouser.html#domousers',
                                                                                          'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUsers._generate_logger': ( 'classes/domouser.html#domousers._generate_logger',
                                                                                                           'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUsers._users_to_domo_user': ( 'classes/domouser.html#domousers._users_to_domo_user',
                                                                                                              'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUsers._users_to_virtual_user': ( 'classes/domouser.html#domousers._users_to_virtual_user',
                                                                                                                 'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUsers.all_users': ( 'classes/domouser.html#domousers.all_users',
                                                                                                    'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUsers.by_email': ( 'classes/domouser.html#domousers.by_email',
                                                                                                   'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUsers.by_id': ( 'classes/domouser.html#domousers.by_id',
                                                                                                'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUsers.util_match_domo_users_to_emails': ( 'classes/domouser.html#domousers.util_match_domo_users_to_emails',
                                                                                                                          'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUsers.util_match_users_obj_to_emails': ( 'classes/domouser.html#domousers.util_match_users_obj_to_emails',
                                                                                                                         'domolibrary/classes/DomoUser.py'),
                                              'domolibrary.classes.DomoUser.DomoUsers.virtual_user_by_subscriber_instance': ( 'classes/domouser.html#domousers.virtual_user_by_subscriber_instance',
                                                                                                                              'domolibrary/classes/DomoUser.py')},
            'domolibrary.client.DomoAuth': { 'domolibrary.client.DomoAuth.DomoAuth': ( 'client/domoauth.html#domoauth',
                                                                                       'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.DomoDeveloperAuth': ( 'client/domoauth.html#domodeveloperauth',
                                                                                                'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.DomoDeveloperAuth.__init__': ( 'client/domoauth.html#domodeveloperauth.__init__',
                                                                                                         'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.DomoDeveloperAuth.generate_auth_header': ( 'client/domoauth.html#domodeveloperauth.generate_auth_header',
                                                                                                                     'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.DomoDeveloperAuth.get_auth_token': ( 'client/domoauth.html#domodeveloperauth.get_auth_token',
                                                                                                               'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.DomoErrror': ( 'client/domoauth.html#domoerrror',
                                                                                         'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.DomoErrror.__init__': ( 'client/domoauth.html#domoerrror.__init__',
                                                                                                  'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.DomoFullAuth': ( 'client/domoauth.html#domofullauth',
                                                                                           'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.DomoFullAuth.generate_auth_header': ( 'client/domoauth.html#domofullauth.generate_auth_header',
                                                                                                                'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.DomoFullAuth.get_auth_token': ( 'client/domoauth.html#domofullauth.get_auth_token',
                                                                                                          'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.DomoTokenAuth': ( 'client/domoauth.html#domotokenauth',
                                                                                            'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.DomoTokenAuth.generate_auth_header': ( 'client/domoauth.html#domotokenauth.generate_auth_header',
                                                                                                                 'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.DomoTokenAuth.get_auth_token': ( 'client/domoauth.html#domotokenauth.get_auth_token',
                                                                                                           'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.InvalidCredentialsError': ( 'client/domoauth.html#invalidcredentialserror',
                                                                                                      'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.InvalidCredentialsError.__init__': ( 'client/domoauth.html#invalidcredentialserror.__init__',
                                                                                                               'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.InvalidInstanceError': ( 'client/domoauth.html#invalidinstanceerror',
                                                                                                   'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.InvalidInstanceError.__init__': ( 'client/domoauth.html#invalidinstanceerror.__init__',
                                                                                                            'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth._DomoAuth_Optional': ( 'client/domoauth.html#_domoauth_optional',
                                                                                                 'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth._DomoAuth_Optional.generate_auth_header': ( 'client/domoauth.html#_domoauth_optional.generate_auth_header',
                                                                                                                      'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth._DomoAuth_Optional.get_auth_token': ( 'client/domoauth.html#_domoauth_optional.get_auth_token',
                                                                                                                'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth._DomoAuth_Optional.print_is_token': ( 'client/domoauth.html#_domoauth_optional.print_is_token',
                                                                                                                'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth._DomoAuth_Required': ( 'client/domoauth.html#_domoauth_required',
                                                                                                 'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth._DomoAuth_Required.__post_init__': ( 'client/domoauth.html#_domoauth_required.__post_init__',
                                                                                                               'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth._DomoAuth_Required.set_manual_login': ( 'client/domoauth.html#_domoauth_required.set_manual_login',
                                                                                                                  'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth._DomoDeveloperAuth_Required': ( 'client/domoauth.html#_domodeveloperauth_required',
                                                                                                          'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth._DomoFullAuth_Required': ( 'client/domoauth.html#_domofullauth_required',
                                                                                                     'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth._DomoTokenAuth_Required': ( 'client/domoauth.html#_domotokenauth_required',
                                                                                                      'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.get_developer_auth': ( 'client/domoauth.html#get_developer_auth',
                                                                                                 'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.get_full_auth': ( 'client/domoauth.html#get_full_auth',
                                                                                            'domolibrary/client/DomoAuth.py'),
                                             'domolibrary.client.DomoAuth.test_access_token': ( 'client/domoauth.html#test_access_token',
                                                                                                'domolibrary/client/DomoAuth.py')},
            'domolibrary.client.Logger': { 'domolibrary.client.Logger.Logger': ( 'client/logger.html#logger',
                                                                                 'domolibrary/client/Logger.py'),
                                           'domolibrary.client.Logger.Logger.__init__': ( 'client/logger.html#logger.__init__',
                                                                                          'domolibrary/client/Logger.py'),
                                           'domolibrary.client.Logger.Logger._add_crumb': ( 'client/logger.html#logger._add_crumb',
                                                                                            'domolibrary/client/Logger.py'),
                                           'domolibrary.client.Logger.Logger._add_log': ( 'client/logger.html#logger._add_log',
                                                                                          'domolibrary/client/Logger.py'),
                                           'domolibrary.client.Logger.Logger._get_traceback': ( 'client/logger.html#logger._get_traceback',
                                                                                                'domolibrary/client/Logger.py'),
                                           'domolibrary.client.Logger.Logger._get_traceback_details': ( 'client/logger.html#logger._get_traceback_details',
                                                                                                        'domolibrary/client/Logger.py'),
                                           'domolibrary.client.Logger.Logger._remove_crumb': ( 'client/logger.html#logger._remove_crumb',
                                                                                               'domolibrary/client/Logger.py'),
                                           'domolibrary.client.Logger.Logger.log_error': ( 'client/logger.html#logger.log_error',
                                                                                           'domolibrary/client/Logger.py'),
                                           'domolibrary.client.Logger.Logger.log_info': ( 'client/logger.html#logger.log_info',
                                                                                          'domolibrary/client/Logger.py'),
                                           'domolibrary.client.Logger.Logger.log_warning': ( 'client/logger.html#logger.log_warning',
                                                                                             'domolibrary/client/Logger.py'),
                                           'domolibrary.client.Logger.Logger.output_log': ( 'client/logger.html#logger.output_log',
                                                                                            'domolibrary/client/Logger.py'),
                                           'domolibrary.client.Logger.TracebackDetails': ( 'client/logger.html#tracebackdetails',
                                                                                           'domolibrary/client/Logger.py')},
            'domolibrary.client.ResponseGetData': { 'domolibrary.client.ResponseGetData.ResponseGetData': ( 'client/responsegetdata.html#responsegetdata',
                                                                                                            'domolibrary/client/ResponseGetData.py'),
                                                    'domolibrary.client.ResponseGetData.ResponseGetData._from_aiohttp_response': ( 'client/responsegetdata.html#responsegetdata._from_aiohttp_response',
                                                                                                                                   'domolibrary/client/ResponseGetData.py'),
                                                    'domolibrary.client.ResponseGetData.ResponseGetData._from_httpx_response': ( 'client/responsegetdata.html#responsegetdata._from_httpx_response',
                                                                                                                                 'domolibrary/client/ResponseGetData.py'),
                                                    'domolibrary.client.ResponseGetData.ResponseGetData._from_looper': ( 'client/responsegetdata.html#responsegetdata._from_looper',
                                                                                                                         'domolibrary/client/ResponseGetData.py'),
                                                    'domolibrary.client.ResponseGetData.ResponseGetData._from_requests_response': ( 'client/responsegetdata.html#responsegetdata._from_requests_response',
                                                                                                                                    'domolibrary/client/ResponseGetData.py'),
                                                    'domolibrary.client.ResponseGetData.ResponseGetData.set_response': ( 'client/responsegetdata.html#responsegetdata.set_response',
                                                                                                                         'domolibrary/client/ResponseGetData.py'),
                                                    'domolibrary.client.ResponseGetData._read_stream': ( 'client/responsegetdata.html#_read_stream',
                                                                                                         'domolibrary/client/ResponseGetData.py'),
                                                    'domolibrary.client.ResponseGetData._write_stream': ( 'client/responsegetdata.html#_write_stream',
                                                                                                          'domolibrary/client/ResponseGetData.py')},
            'domolibrary.client.get_data': { 'domolibrary.client.get_data.LooperError': ( 'client/get_data.html#loopererror',
                                                                                          'domolibrary/client/get_data.py'),
                                             'domolibrary.client.get_data.LooperError.__init__': ( 'client/get_data.html#loopererror.__init__',
                                                                                                   'domolibrary/client/get_data.py'),
                                             'domolibrary.client.get_data.get_data': ( 'client/get_data.html#get_data',
                                                                                       'domolibrary/client/get_data.py'),
                                             'domolibrary.client.get_data.get_data_aiohttp': ( 'client/get_data.html#get_data_aiohttp',
                                                                                               'domolibrary/client/get_data.py'),
                                             'domolibrary.client.get_data.looper': ( 'client/get_data.html#looper',
                                                                                     'domolibrary/client/get_data.py')},
            'domolibrary.integrations.DomoJupyter': { 'domolibrary.integrations.DomoJupyter.DojoAccount_InstanceAuth': ( 'integrations/domojupyter.html#dojoaccount_instanceauth',
                                                                                                                         'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.DojoAccount_InstanceAuth.__post_init__': ( 'integrations/domojupyter.html#dojoaccount_instanceauth.__post_init__',
                                                                                                                                       'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.DojoAccount_InstanceAuth._clean_account_admin_accounts': ( 'integrations/domojupyter.html#dojoaccount_instanceauth._clean_account_admin_accounts',
                                                                                                                                                       'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.DojoAccount_InstanceAuth._generate_auth': ( 'integrations/domojupyter.html#dojoaccount_instanceauth._generate_auth',
                                                                                                                                        'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.DojoAccount_InstanceAuth._set_display_name': ( 'integrations/domojupyter.html#dojoaccount_instanceauth._set_display_name',
                                                                                                                                           'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.DojoAccount_InstanceAuth._test_regex_mask': ( 'integrations/domojupyter.html#dojoaccount_instanceauth._test_regex_mask',
                                                                                                                                          'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.DojoAccount_InstanceAuth.generate_auth_ls': ( 'integrations/domojupyter.html#dojoaccount_instanceauth.generate_auth_ls',
                                                                                                                                          'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.DojoAccount_InstanceAuth.get_domo_instance_auth_account': ( 'integrations/domojupyter.html#dojoaccount_instanceauth.get_domo_instance_auth_account',
                                                                                                                                                        'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GenerateAuth_CredentialsNotProvided': ( 'integrations/domojupyter.html#generateauth_credentialsnotprovided',
                                                                                                                                    'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GenerateAuth_CredentialsNotProvided.__init__': ( 'integrations/domojupyter.html#generateauth_credentialsnotprovided.__init__',
                                                                                                                                             'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GenerateAuth_InvalidDomoInstanceList': ( 'integrations/domojupyter.html#generateauth_invaliddomoinstancelist',
                                                                                                                                     'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GenerateAuth_InvalidDomoInstanceList.__init__': ( 'integrations/domojupyter.html#generateauth_invaliddomoinstancelist.__init__',
                                                                                                                                              'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GetDomains_Query_AuthMatch_Error': ( 'integrations/domojupyter.html#getdomains_query_authmatch_error',
                                                                                                                                 'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GetDomains_Query_AuthMatch_Error.__init__': ( 'integrations/domojupyter.html#getdomains_query_authmatch_error.__init__',
                                                                                                                                          'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GetDomains_Query_Exception_PW_Col_Error': ( 'integrations/domojupyter.html#getdomains_query_exception_pw_col_error',
                                                                                                                                        'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GetDomains_Query_Exception_PW_Col_Error.__init__': ( 'integrations/domojupyter.html#getdomains_query_exception_pw_col_error.__init__',
                                                                                                                                                 'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GetInstanceConfig': ( 'integrations/domojupyter.html#getinstanceconfig',
                                                                                                                  'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GetInstanceConfig.__init__': ( 'integrations/domojupyter.html#getinstanceconfig.__init__',
                                                                                                                           'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GetInstanceConfig._retrieve_company_ds': ( 'integrations/domojupyter.html#getinstanceconfig._retrieve_company_ds',
                                                                                                                                       'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GetInstanceConfig.get_domains_with_global_config_auth': ( 'integrations/domojupyter.html#getinstanceconfig.get_domains_with_global_config_auth',
                                                                                                                                                      'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GetInstanceConfig.get_domains_with_instance_auth': ( 'integrations/domojupyter.html#getinstanceconfig.get_domains_with_instance_auth',
                                                                                                                                                 'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GetJupyter_ErrorRetrievingAccount': ( 'integrations/domojupyter.html#getjupyter_errorretrievingaccount',
                                                                                                                                  'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GetJupyter_ErrorRetrievingAccount.__init__': ( 'integrations/domojupyter.html#getjupyter_errorretrievingaccount.__init__',
                                                                                                                                           'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GetJupyter_ErrorRetrievingAccountProperty': ( 'integrations/domojupyter.html#getjupyter_errorretrievingaccountproperty',
                                                                                                                                          'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.GetJupyter_ErrorRetrievingAccountProperty.__init__': ( 'integrations/domojupyter.html#getjupyter_errorretrievingaccountproperty.__init__',
                                                                                                                                                   'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.InvalidAccountNameError': ( 'integrations/domojupyter.html#invalidaccountnameerror',
                                                                                                                        'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.InvalidAccountNameError.__init__': ( 'integrations/domojupyter.html#invalidaccountnameerror.__init__',
                                                                                                                                 'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.InvalidAccountTypeError': ( 'integrations/domojupyter.html#invalidaccounttypeerror',
                                                                                                                        'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.InvalidAccountTypeError.__init__': ( 'integrations/domojupyter.html#invalidaccounttypeerror.__init__',
                                                                                                                                 'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.NoConfigCompanyError': ( 'integrations/domojupyter.html#noconfigcompanyerror',
                                                                                                                     'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.NoConfigCompanyError.__init__': ( 'integrations/domojupyter.html#noconfigcompanyerror.__init__',
                                                                                                                              'domolibrary/integrations/DomoJupyter.py'),
                                                      'domolibrary.integrations.DomoJupyter.get_jupyter_account': ( 'integrations/domojupyter.html#get_jupyter_account',
                                                                                                                    'domolibrary/integrations/DomoJupyter.py')},
            'domolibrary.routes.account': { 'domolibrary.routes.account.AccountConfig_InvalidDataProvider': ( 'routes/account.html#accountconfig_invaliddataprovider',
                                                                                                              'domolibrary/routes/account.py'),
                                            'domolibrary.routes.account.AccountConfig_InvalidDataProvider.__init__': ( 'routes/account.html#accountconfig_invaliddataprovider.__init__',
                                                                                                                       'domolibrary/routes/account.py'),
                                            'domolibrary.routes.account.create_account': ( 'routes/account.html#create_account',
                                                                                           'domolibrary/routes/account.py'),
                                            'domolibrary.routes.account.delete_account': ( 'routes/account.html#delete_account',
                                                                                           'domolibrary/routes/account.py'),
                                            'domolibrary.routes.account.get_account_config': ( 'routes/account.html#get_account_config',
                                                                                               'domolibrary/routes/account.py'),
                                            'domolibrary.routes.account.get_account_from_id': ( 'routes/account.html#get_account_from_id',
                                                                                                'domolibrary/routes/account.py'),
                                            'domolibrary.routes.account.get_accounts': ( 'routes/account.html#get_accounts',
                                                                                         'domolibrary/routes/account.py'),
                                            'domolibrary.routes.account.update_account_config': ( 'routes/account.html#update_account_config',
                                                                                                  'domolibrary/routes/account.py'),
                                            'domolibrary.routes.account.update_account_name': ( 'routes/account.html#update_account_name',
                                                                                                'domolibrary/routes/account.py')},
            'domolibrary.routes.activity_log': { 'domolibrary.routes.activity_log.get_activity_log_object_types': ( 'routes/activity_log.html#get_activity_log_object_types',
                                                                                                                    'domolibrary/routes/activity_log.py'),
                                                 'domolibrary.routes.activity_log.search_activity_log': ( 'routes/activity_log.html#search_activity_log',
                                                                                                          'domolibrary/routes/activity_log.py')},
            'domolibrary.routes.bootstrap': { 'domolibrary.routes.bootstrap.GetBootstrap_InvalidAuthMethod': ( 'routes/bootstrap.html#getbootstrap_invalidauthmethod',
                                                                                                               'domolibrary/routes/bootstrap.py'),
                                              'domolibrary.routes.bootstrap.GetBootstrap_InvalidAuthMethod.__init__': ( 'routes/bootstrap.html#getbootstrap_invalidauthmethod.__init__',
                                                                                                                        'domolibrary/routes/bootstrap.py'),
                                              'domolibrary.routes.bootstrap.get_bootstrap': ( 'routes/bootstrap.html#get_bootstrap',
                                                                                              'domolibrary/routes/bootstrap.py'),
                                              'domolibrary.routes.bootstrap.get_bootstrap_features': ( 'routes/bootstrap.html#get_bootstrap_features',
                                                                                                       'domolibrary/routes/bootstrap.py'),
                                              'domolibrary.routes.bootstrap.get_bootstrap_pages': ( 'routes/bootstrap.html#get_bootstrap_pages',
                                                                                                    'domolibrary/routes/bootstrap.py')},
            'domolibrary.routes.dataset': { 'domolibrary.routes.dataset.DatasetNotFoundError': ( 'routes/dataset.html#datasetnotfounderror',
                                                                                                 'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.DatasetNotFoundError.__init__': ( 'routes/dataset.html#datasetnotfounderror.__init__',
                                                                                                          'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.QueryRequestError': ( 'routes/dataset.html#queryrequesterror',
                                                                                              'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.QueryRequestError.__init__': ( 'routes/dataset.html#queryrequesterror.__init__',
                                                                                                       'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.UploadDataError': ( 'routes/dataset.html#uploaddataerror',
                                                                                            'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.UploadDataError.__init__': ( 'routes/dataset.html#uploaddataerror.__init__',
                                                                                                     'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.create': ( 'routes/dataset.html#create',
                                                                                   'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.generate_create_dataset_body': ( 'routes/dataset.html#generate_create_dataset_body',
                                                                                                         'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.generate_list_partitions_body': ( 'routes/dataset.html#generate_list_partitions_body',
                                                                                                          'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.get_dataset_by_id': ( 'routes/dataset.html#get_dataset_by_id',
                                                                                              'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.get_schema': ( 'routes/dataset.html#get_schema',
                                                                                       'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.index_dataset': ( 'routes/dataset.html#index_dataset',
                                                                                          'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.index_status': ( 'routes/dataset.html#index_status',
                                                                                         'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.list_partitions': ( 'routes/dataset.html#list_partitions',
                                                                                            'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.query_dataset_private': ( 'routes/dataset.html#query_dataset_private',
                                                                                                  'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.query_dataset_public': ( 'routes/dataset.html#query_dataset_public',
                                                                                                 'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.set_dataset_tags': ( 'routes/dataset.html#set_dataset_tags',
                                                                                             'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.upload_dataset_stage_1': ( 'routes/dataset.html#upload_dataset_stage_1',
                                                                                                   'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.upload_dataset_stage_2_df': ( 'routes/dataset.html#upload_dataset_stage_2_df',
                                                                                                      'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.upload_dataset_stage_2_file': ( 'routes/dataset.html#upload_dataset_stage_2_file',
                                                                                                        'domolibrary/routes/dataset.py'),
                                            'domolibrary.routes.dataset.upload_dataset_stage_3': ( 'routes/dataset.html#upload_dataset_stage_3',
                                                                                                   'domolibrary/routes/dataset.py')},
            'domolibrary.routes.user': { 'domolibrary.routes.user.create_user': ( 'routes/user.html#create_user',
                                                                                  'domolibrary/routes/user.py'),
                                         'domolibrary.routes.user.generate_search_users_body_by_email': ( 'routes/user.html#generate_search_users_body_by_email',
                                                                                                          'domolibrary/routes/user.py'),
                                         'domolibrary.routes.user.generate_search_users_body_by_id': ( 'routes/user.html#generate_search_users_body_by_id',
                                                                                                       'domolibrary/routes/user.py'),
                                         'domolibrary.routes.user.get_all_users': ( 'routes/user.html#get_all_users',
                                                                                    'domolibrary/routes/user.py'),
                                         'domolibrary.routes.user.process_v1_search_users': ( 'routes/user.html#process_v1_search_users',
                                                                                              'domolibrary/routes/user.py'),
                                         'domolibrary.routes.user.request_password_reset': ( 'routes/user.html#request_password_reset',
                                                                                             'domolibrary/routes/user.py'),
                                         'domolibrary.routes.user.reset_password': ( 'routes/user.html#reset_password',
                                                                                     'domolibrary/routes/user.py'),
                                         'domolibrary.routes.user.search_users': ( 'routes/user.html#search_users',
                                                                                   'domolibrary/routes/user.py'),
                                         'domolibrary.routes.user.search_virtual_user_by_subscriber_instance': ( 'routes/user.html#search_virtual_user_by_subscriber_instance',
                                                                                                                 'domolibrary/routes/user.py'),
                                         'domolibrary.routes.user.set_user_landing_page': ( 'routes/user.html#set_user_landing_page',
                                                                                            'domolibrary/routes/user.py')},
            'domolibrary.utils.DictDot': { 'domolibrary.utils.DictDot.DictDot': ( 'utils/dictdot.html#dictdot',
                                                                                  'domolibrary/utils/DictDot.py'),
                                           'domolibrary.utils.DictDot.DictDot.__getattr__': ( 'utils/dictdot.html#dictdot.__getattr__',
                                                                                              'domolibrary/utils/DictDot.py'),
                                           'domolibrary.utils.DictDot.DictDot.__init__': ( 'utils/dictdot.html#dictdot.__init__',
                                                                                           'domolibrary/utils/DictDot.py')},
            'domolibrary.utils.convert': { 'domolibrary.utils.convert.convert_datetime_to_epoch_millisecond': ( 'utils/convert.html#convert_datetime_to_epoch_millisecond',
                                                                                                                'domolibrary/utils/convert.py'),
                                           'domolibrary.utils.convert.convert_epoch_millisecond_to_datetime': ( 'utils/convert.html#convert_epoch_millisecond_to_datetime',
                                                                                                                'domolibrary/utils/convert.py'),
                                           'domolibrary.utils.convert.convert_snake_to_pascal': ( 'utils/convert.html#convert_snake_to_pascal',
                                                                                                  'domolibrary/utils/convert.py')}}}



import datetime as dt


def convert_epoch_millisecond_to_datetime(epoch: int):
    '''convert Epoch time with miliseconds to Date time'''
    return dt.datetime.fromtimestamp(epoch / 1000.0)


def convert_datetime_to_epoch_millisecond(datetime: dt.datetime):
    '''convert DateTime to Epoch time with Miliseconds'''
    return int(datetime.timestamp() * 1000)


def convert_snake_to_pascal(str):
    clean_str = str.replace("_", " ").title().replace(" ", "")
    return clean_str[0].lower()+clean_str[1:]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/convert.ipynb.

# %% auto 0
__all__ = ['convert_epoch_millisecond_to_datetime', 'convert_datetime_to_epoch_millisecond', 'convert_snake_to_pascal']

# %% ../../nbs/utils/convert.ipynb 2
import datetime as dt

# %% ../../nbs/utils/convert.ipynb 3
import fastcore.test as fctest

# %% ../../nbs/utils/convert.ipynb 4
def convert_epoch_millisecond_to_datetime(epoch: int):
    '''convert Epoch time with miliseconds to Date time'''
    return dt.datetime.fromtimestamp(epoch / 1000.0)

# %% ../../nbs/utils/convert.ipynb 5
def convert_datetime_to_epoch_millisecond(datetime: dt.datetime):
    '''convert DateTime to Epoch time with Miliseconds'''
    return int(datetime.timestamp() * 1000)

# %% ../../nbs/utils/convert.ipynb 8
def convert_snake_to_pascal(str):
    """converts 'snake_case_str' to 'snakeCaseStr'"""
    
    clean_str = str.replace("_", " ").title().replace(" ", "")
    return clean_str[0].lower()+clean_str[1:]



from configparser import ConfigParser

import setuptools
from pkg_resources import parse_version

assert parse_version(setuptools.__version__) >= parse_version('36.2')

# note: all settings are in settings.ini; edit there, not here
config = ConfigParser(delimiters=['='])
config.read('settings.ini')
cfg = config['DEFAULT']

cfg_keys = 'version description keywords author author_email'.split()
expected = cfg_keys + \
    "lib_name user branch license status min_python audience language".split()
for o in expected:
    assert o in cfg, "missing expected setting: {}".format(o)
setup_cfg = {o: cfg[o] for o in cfg_keys}

licenses = {
    'apache2': ('Apache Software License 2.0', 'OSI Approved :: Apache Software License'),
    'mit': ('MIT License', 'OSI Approved :: MIT License'),
    'gpl2': ('GNU General Public License v2', 'OSI Approved :: GNU General Public License v2 (GPLv2)'),
    'gpl3': ('GNU General Public License v3', 'OSI Approved :: GNU General Public License v3 (GPLv3)'),
    'bsd3': ('BSD License', 'OSI Approved :: BSD License'),
}
statuses = ['1 - Planning', '2 - Pre-Alpha', '3 - Alpha',
            '4 - Beta', '5 - Production/Stable', '6 - Mature', '7 - Inactive']
py_versions = '3.6 3.7 3.8 3.9 3.10'.split()

requirements = cfg.get('requirements', '').split()
if cfg.get('pip_requirements'):
    requirements += cfg.get('pip_requirements', '').split()
min_python = cfg['min_python']
lic = licenses.get(cfg['license'].lower(), (cfg['license'], None))
dev_requirements = (cfg.get('dev_requirements') or '').split()

setuptools.setup(
    name=cfg['lib_name'],
    license=lic[0],
    classifiers=[
        'Development Status :: ' + statuses[int(cfg['status'])],
        'Intended Audience :: ' + cfg['audience'].title(),
        'Natural Language :: ' + cfg['language'].title(),
    ] + ['Programming Language :: Python :: '+o for o in py_versions[py_versions.index(min_python):]] + (['License :: ' + lic[1]] if lic[1] else []),
    url=cfg['git_url'],
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=requirements,
    extras_require={'dev': dev_requirements},
    dependency_links=cfg.get('dep_links', '').split(),
    python_requires='>=' + cfg['min_python'],
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    zip_safe=False,
    entry_points={
        'console_scripts': cfg.get('console_scripts', '').split(),
        'nbdev': [f'{cfg.get("lib_path")}={cfg.get("lib_path")}._modidx:d']
    },
    **setup_cfg)



import aiohttp

from .get_data import get_data
from ..DomoAuth import DomoFullAuth
from ...utils.ResponseGetData import ResponseGetData


async def get_stream_definition(full_auth: DomoFullAuth, stream_id: str,
                                session: aiohttp.ClientSession = None, debug: bool = False) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/data/v1/streams/{stream_id}'

    if debug:
        print(url)

    res = await get_data(
        auth=full_auth,
        url=url,
        method='GET',
        session=session,
        debug=debug,
    )
    return res


async def update_stream(full_auth: DomoFullAuth, stream_id: str,
                        body: dict,
                        session: aiohttp.ClientSession = None,
                        debug: bool = False) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/data/v1/streams/{stream_id}'

    if debug:
        print(url)

    res = await get_data(
        auth=full_auth,
        url=url,
        body=body,
        method='PUT',
        session=session,
        debug=debug,
    )
    return res


async def create_stream(full_auth: DomoFullAuth,
                        body: dict,
                        session: aiohttp.ClientSession = None,
                        debug: bool = False) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/data/v1/streams'

    if debug:
        print(url)

    res = await get_data(
        auth=full_auth,
        url=url,
        body=body,
        method='POST',
        session=session,
        debug=debug,
    )
    return res



import asyncio
import aiohttp
from dataclasses import dataclass, field

from .routes import page_routes
from .DomoAuth import DomoDeveloperAuth, DomoFullAuth
from ..utils.DictDot import DictDot
from ..utils.Base import Base


@dataclass
class DomoPage(Base):
    id: str
    title: str = None
    parent_page_id: str = None
    domo_instance: str = None
    full_auth: DomoFullAuth = None
    owners: list = field(default_factory=list)
    cards: list = field(default_factory=list)
    collections: list = field(default_factory=list)

    def __post_init__(self, full_auth=None):
        Base().__init__()

        if full_auth:
            self.domo_instance = full_auth.domo_instance
            self.full_auth = full_auth
        # self.Definition = CardDefinition(self)

    def display_url(self):
        return f'https://{self.domo_instance}.domo.com/page/{self.id}'

    @classmethod
    async def get_from_id(cls, page_id: str, full_auth: DomoFullAuth, debug: bool = False):
        import Library.DomoClasses.DomoCard as dc
        res = await page_routes.get_page_by_id(full_auth=full_auth, page_id=page_id, debug=debug)

        if res.status == 200:
            dd = DictDot(res.response)

            pg = cls(
                id=dd.id,
                domo_instance=full_auth.domo_instance,
                title=dd.title,
                parent_page_id=dd.page.parentPageId,
                owners=dd.page.owners,
                collections=dd.collections
            )

            pg.cards = await asyncio.gather(
                *[dc.DomoCard.get_from_id(id=card.id, full_auth=full_auth) for card in dd.cards])

            return pg

    @classmethod
    async def get_cards(cls, full_auth, page_id, debug: bool = False, session: aiohttp.ClientSession = None):
        try:
            import Library.DomoClasses.DomoCard as dc
            close_session = False if session else True

            if not session:
                session = aiohttp.ClientSession()

            res = await page_routes.get_page_definition(full_auth=full_auth, page_id=page_id, debug=debug, session=session)

            if res.status == 200:
                json = res.response

                card_list = [dc.DomoCard(id=card.get(
                    'id'), full_auth=full_auth) for card in json.get('cards')]

                return card_list

            else:
                return None

        finally:
            if close_session:
                await session.close()

    async def get_datasets(full_auth, page_id, debug: bool = False, session: aiohttp.ClientSession = None):
        try:
            import Library.DomoClasses.DomoDataset as dmds
            close_session = False if session else True

            if not session:
                session = aiohttp.ClientSession()

            res = await page_routes.get_page_definition(full_auth=full_auth, page_id=page_id, debug=debug, session=session)

            if res.status == 200:
                json = res.response

                dataset_ls = [card.get('datasources')
                              for card in json.get('cards')]

                return [dmds.DomoDataset(id=ds.get('dataSourceId'), full_auth=full_auth) for ds_ls in dataset_ls for ds in ds_ls]

            else:
                return None

        finally:
            if close_session:
                await session.close()
